import logging
import os
from contextlib import contextmanager, nullcontext
from operator import attrgetter, itemgetter
from typing import Any, Dict, List, NamedTuple, Sequence, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pint
from IPython.display import HTML
from matplotlib.patches import PathPatch
from matplotlib.path import Path
from shapely import affinity
from shapely.geometry import Point

from ..em import ureg
from ..finite_volume.mesh import Mesh
from ..finite_volume.util import get_oriented_boundary
from ..geometry import ensure_unique
from .layer import Layer
from .meshing import generate_mesh
from .polygon import Polygon

logger = logging.getLogger(__name__)


class TerminalInfo(NamedTuple):
    """A containter for information about a single current terminal.

    Args:
        name: The terminal's name.
        site_indices: An array of indices for mesh sites making up the terminal.
        edge_indices: An array of indices for mesh edges making up the terminal.
        boundary_edge_indices: An array of indices of mesh boundary edges making
            up the terminal.
        length: The length of the terminal in physical units.
    """

    name: str
    site_indices: Sequence[int]
    edge_indices: Sequence[int]
    boundary_edge_indices: Sequence[int]
    length: float


class Device:
    """An object representing a thin film superconducting device.

    Args:
        name: Name of the device.
        layer: The superconducting ``Layer``.
        film: The ``Polygon`` representing the superconducting film.
        holes: ``Polygons`` representing holes in the superconducting film.
        abstract_regions: ``Polygons`` representing abstract regions in a device.
            Abstract regions are areas that can be meshed, but need not correspond
            to any physical structres in the device.
        terminals: A sequence of ``Polygons`` representing the current terminals.
            Any points that are on the boundary of the mesh and lie inside a
            terminal will have current source/sink boundary conditions.
        probe_points: A shape ``(n, 2)`` sequence of floats, with each row
            representing the ``(x, y)`` position of a voltage probe.
        length_units: Distance units for the coordinate system.
    """

    ureg = ureg

    def __init__(
        self,
        name: str,
        *,
        layer: Layer,
        film: Polygon,
        holes: Union[List[Polygon], None] = None,
        abstract_regions: Union[List[Polygon], None] = None,
        terminals: Union[List[Polygon], None] = None,
        probe_points: Sequence[float] = None,
        length_units: str = "um",
    ):
        self.name = name
        self.layer = layer
        self.film = film
        self.holes = holes or []
        self.abstract_regions = abstract_regions or []
        self.terminals = tuple(terminals or [])
        names = set()
        for terminal in self.terminals:
            terminal.mesh = False
            if terminal.name is None or terminal.name in names:
                raise ValueError("All terminals must have a unique name")
            names.add(terminal.name)

        for polygon in [self.film] + self.holes:
            if not polygon.is_valid:
                raise ValueError("Invalid Polygon: {polygon!r}.")

        if probe_points is not None:
            probe_points = np.asarray(probe_points).squeeze()
            if probe_points.ndim != 2 or probe_points.shape[1] != 2:
                raise ValueError(
                    f"Probe points must have shape (n, 2), "
                    f"got {probe_points.shape}."
                )
            if not self.contains_points(probe_points).all():
                raise ValueError("All probe points must lie within the film.")
        self.probe_points = probe_points

        # Make units a "read-only" attribute.
        # It should never be changed after instantiation.
        self._length_units = length_units

        self.mesh = None
        self.mesh_info = None

    @property
    def length_units(self) -> str:
        """Length units used for the device geometry."""
        return self._length_units

    @property
    def coherence_length(self) -> pint.Quantity:
        """Ginzburg-Landau coherence length, :math:`\\xi`"""
        return self.layer.coherence_length * ureg(self.length_units)

    @property
    def london_lambda(self) -> pint.Quantity:
        """London penetration depth, :math:`\\lambda`"""
        return self.layer.london_lambda * ureg(self.length_units)

    @property
    def thickness(self) -> pint.Quantity:
        """Film thickness, :math:`d`"""
        return self.layer.thickness * ureg(self.length_units)

    @property
    def Lambda(self) -> pint.Quantity:
        """Effective magnetic penetration depth, :math:`\\Lambda=\\lambda^2/d`."""
        return self.london_lambda**2 / self.thickness

    @property
    def conductivity(self) -> Union[pint.Quantity, None]:
        """Film normal state conductivity, :math:`\\sigma`"""
        if self.layer.conductivity is None:
            return None
        return self.layer.conductivity * ureg(f"siemens / {self.length_units}")

    @property
    def kappa(self) -> float:
        """The Ginzburg-Landau parameter, :math:`\\kappa=\\lambda/\\xi`."""
        return (self.london_lambda / self.coherence_length).magnitude

    @property
    def Bc2(self) -> pint.Quantity:
        """Upper critical field, :math:`B_{c2}=\\Phi_0/(2\\pi\\xi^2)`."""
        return (
            ureg("Phi_0") / (2 * np.pi * self.coherence_length**2)
        ).to_base_units()

    @property
    def A0(self) -> pint.Quantity:
        """Scale for the magnetic vector potential, :math:`A_0=\\xi B_{c2}`."""
        return (self.Bc2 * self.coherence_length).to_base_units()

    @property
    def K0(self) -> pint.Quantity:
        """Sheet current density scale (dimensions of current / length),
        :math:`K_0=4\\xi B_{c2}/(\\mu_0\\Lambda)`.
        """
        K0 = 4 * self.coherence_length * self.Bc2 / (ureg("mu_0") * self.Lambda)
        return K0.to_base_units()

    def tau0(self, conductivity: Union[pint.Quantity, None] = None) -> pint.Quantity:
        """Time scale, :math:`\\tau_0=\\mu_0\\sigma\\lambda^2`.

        Args:
            conductivity: The normal state conductivity of the film, which defaults
                to ``device.layer.conductivity``.

        Returns:
            The time scale, :math:`\\tau_0=\\mu_0\\sigma\\lambda^2
        """
        if conductivity is None:
            conductivity = self.conductivity
        if conductivity is None:
            raise ValueError(
                "The time scale tau0 requires the normal state"
                " conductivity to be defined."
            )
        return (ureg("mu_0") * conductivity * self.london_lambda**2).to("seconds")

    def V0(self, conductivity: Union[pint.Quantity, None] = None) -> pint.Quantity:
        """Electric potential scale, :math:`\\V_0=\\xi J_0/\\sigma`.

        Args:
            conductivity: The normal state conductivity of the film, which defaults
                to ``device.layer.conductivity``.

        Returns:
            The electric potential scale, :math:`\\V_0=\\xi J_0/\\sigma`
        """
        if conductivity is None:
            conductivity = self.conductivity
        if conductivity is None:
            raise ValueError(
                "The electric potential scale V_0 requires the normal state"
                " conductivity to be defined."
            )
        J0 = self.K0 / self.thickness
        return (self.coherence_length * J0 / conductivity).to("volts")

    def terminal_info(self) -> Tuple[TerminalInfo, ...]:
        """Returns a tuple of ``TerminalInfo`` objects,
        one for each current terminal in the device.
        """
        xi = self.layer.coherence_length
        mesh = self.mesh
        sites = self.points
        edge_positions = xi * mesh.edge_mesh.centers
        ix_boundary = mesh.edge_mesh.boundary_edge_indices
        edge_lengths = self.edge_lengths[ix_boundary]
        boundary_edge_positions = edge_positions[ix_boundary]
        info = []
        for terminal in self.terminals:
            # Index into self.points
            sites_index = np.intersect1d(
                terminal.contains_points(sites, index=True), mesh.boundary_indices
            )
            # Index into self.edges
            edges_index = np.intersect1d(
                terminal.contains_points(edge_positions, index=True), ix_boundary
            )
            # Index into self.edges[mesh.edge_mesh.boundary_edge_indices]
            boundary_edges_index = terminal.contains_points(
                boundary_edge_positions, index=True
            )
            length = edge_lengths[boundary_edges_index].sum()
            info.append(
                TerminalInfo(
                    terminal.name,
                    sites_index,
                    edges_index,
                    boundary_edges_index,
                    length,
                )
            )
        return tuple(sorted(info, key=attrgetter("length")))

    @property
    def polygons(self) -> Tuple[Polygon, ...]:
        """Tuple of all ``Polygons`` in the ``Device``."""
        return (
            (self.film,)
            + tuple(self.holes)
            + tuple(self.abstract_regions)
            + self.terminals
        )

    @property
    def points(self) -> Union[np.ndarray, None]:
        """The mesh vertex coordinates in ``length_units``
        (shape ``(n, 2)``, type ``float``).
        """
        if self.mesh is None:
            return None
        return self.mesh.sites * self.coherence_length.magnitude

    @property
    def triangles(self) -> Union[np.ndarray, None]:
        """Mesh triangle indices (shape ``(m, 3)``, type ``int``)."""
        if self.mesh is None:
            return None
        return self.mesh.elements

    @property
    def edges(self) -> Union[np.ndarray, None]:
        """Mesh edge indices (shape ``(p, 2)``, type ``int``)."""
        if self.mesh is None:
            return None
        return self.mesh.edge_mesh.edges

    @property
    def edge_lengths(self) -> Union[np.ndarray, None]:
        """An array of the mesh vertex-to-vertex distances."""
        if self.mesh is None:
            return None
        return self.mesh.edge_mesh.edge_lengths * self.coherence_length.magnitude

    @property
    def areas(self) -> Union[np.ndarray, None]:
        """An array of the mesh Voronoi cell areas."""
        if self.mesh is None:
            return None
        return self.mesh.areas * self.coherence_length.magnitude**2

    @property
    def probe_point_indices(self) -> Union[List[int], None]:
        """A list of the mesh site indices for the probe points."""
        if self.mesh is None or self.probe_points is None:
            return None
        xi = self.coherence_length.magnitude
        return [self.mesh.closest_site(xy) for xy in self.probe_points / xi]

    def boundary_sites(self) -> Union[Dict[str, np.ndarray], None]:
        """Returns a dict of ``{polygon_name: boundary_indices}``, where ``boundary_indices``
        is an integer array of site indices for mesh sites on the boundary of each polygon.

        The length of the returned dictionary will be the number of holes in the device
        plus one.

        Returns:
            ``{polygon_name: boundary_indices}``
        """
        if self.mesh is None:
            return None
        polygons = [self.film] + list(self.holes)
        points = self.points
        edge_mesh = self.mesh.edge_mesh
        boundary_edges = edge_mesh.edges[edge_mesh.boundary_edge_indices]
        boundary = {}
        for polygon in polygons:
            on_boundary = np.logical_and(
                polygon.on_boundary(points[boundary_edges[:, 0]], radius=1e-6),
                polygon.on_boundary(points[boundary_edges[:, 1]], radius=1e-6),
            )
            boundary_sites = get_oriented_boundary(points, boundary_edges[on_boundary])
            assert len(boundary_sites) == 1, len(boundary_sites)
            boundary[polygon.name] = boundary_sites[0]
        return boundary

    def contains_points(
        self,
        points: np.ndarray,
        index: bool = False,
        radius: float = 0,
    ) -> np.ndarray:
        """Determines whether ``points`` lie within the Device.

        Args:
            points: Shape ``(n, 2)`` array of x, y coordinates.
            index: If True, then return the indices of the points in ``points``
                that lie within the polygon. Otherwise, returns a shape ``(n, )``
                boolean array.
            radius: An additional margin on ``self.path``.
                See :meth:`matplotlib.path.Path.contains_points`.

        Returns:
            If index is True, returns the indices of the points in ``points``
            that lie within the polygon. Otherwise, returns a shape ``(n, )``
            boolean array indicating whether each point lies within the polygon.
        """
        mask = self.film.contains_points(points, radius=radius) & ~np.logical_or.reduce(
            [hole.contains_points(points, radius=-radius) for hole in self.holes]
        )
        if index:
            return np.where(mask)[0]
        return mask

    @property
    def poly_points(self) -> np.ndarray:
        """A shape (n, 2) array of (x, y) coordinates of all Polygons in the Device."""
        points = np.concatenate(
            [self.film.points]
            + [poly.points for poly in self.abstract_regions if poly.mesh],
            axis=0,
        )
        # Remove duplicate points to avoid meshing issues.
        # If you don't do this and there are duplicate points,
        # meshpy.triangle will segfault.
        return ensure_unique(points)

    def copy(self, with_mesh: bool = True) -> "Device":
        """Copy this Device to create a new one.

        Args:
            with_mesh: Whether to copy the mesh.

        Returns:
            A new Device instance, copied from self.
        """
        holes = [hole.copy() for hole in self.holes]
        abstract_regions = [region.copy() for region in self.abstract_regions]
        terminals = [term.copy() for term in self.terminals]
        if self.probe_points is None:
            probe_points = None
        else:
            probe_points = self.probe_points.copy()

        device = Device(
            self.name,
            layer=self.layer.copy(),
            film=self.film.copy(),
            holes=holes,
            abstract_regions=abstract_regions,
            terminals=terminals,
            probe_points=probe_points,
            length_units=self.length_units,
        )
        if with_mesh and self.mesh is not None:
            device.mesh = self.mesh
        return device

    def _warn_if_mesh_exist(self, method: str) -> None:
        if self.mesh is not None:
            message = (
                f"Calling device.{method} on a device whose mesh already exists "
                f"returns a new device with no mesh. Call new_device.make_mesh() "
                f"to generate the mesh for the new device."
            )
            logger.warning(message)

    def scale(
        self, xfact: float = 1, yfact: float = 1, origin: Tuple[float, float] = (0, 0)
    ) -> "Device":
        """Returns a new device with polygons scaled horizontally and/or vertically.

        Negative ``xfact`` (``yfact``) can be used to reflect the device horizontally
        (vertically) about the ``origin``.

        Args:
            xfact: Factor by which to scale the device horizontally.
            yfact: Factor by which to scale the device vertically.
            origin: (x, y) coorindates of the origin.

        Returns:
            The scaled ``Device``.
        """
        if not (
            isinstance(origin, tuple)
            and len(origin) == 2
            and all(isinstance(val, (int, float)) for val in origin)
        ):
            raise TypeError("Origin must be a tuple of floats (x, y).")
        self._warn_if_mesh_exist("scale()")
        device = self.copy(with_mesh=False)
        for polygon in device.polygons:
            polygon.scale(xfact=xfact, yfact=yfact, origin=origin, inplace=True)
        if device.probe_points is not None:
            points = [
                affinity.scale(Point(xy), xfact=xfact, yfact=yfact, origin=origin)
                for xy in device.probe_points
            ]
            device.probe_points = np.concatenate(
                [point.coords for point in points], axis=0
            )
        return device

    def rotate(self, degrees: float, origin: Tuple[float, float] = (0, 0)) -> "Device":
        """Returns a new device with polygons rotated a given amount
        counterclockwise about specified origin.

        Args:
            degrees: The amount by which to rotate the polygons.
            origin: (x, y) coorindates of the origin.

        Returns:
            The rotated ``Device``.
        """
        if not (
            isinstance(origin, tuple)
            and len(origin) == 2
            and all(isinstance(val, (int, float)) for val in origin)
        ):
            raise TypeError("Origin must be a tuple of floats (x, y).")
        self._warn_if_mesh_exist("rotate()")
        device = self.copy(with_mesh=False)
        for polygon in device.polygons:
            polygon.rotate(degrees, origin=origin, inplace=True)
        if self.probe_points is not None:
            points = [
                affinity.rotate(Point(xy), degrees, origin=origin)
                for xy in self.probe_points
            ]
            device.probe_points = np.concatenate(
                [point.coords for point in points], axis=0
            )
        return device

    def translate(
        self,
        dx: float = 0,
        dy: float = 0,
        dz: float = 0,
        inplace: bool = False,
    ) -> "Device":
        """Translates the device polygons, layer, and mesh in space by a given amount.

        Args:
            dx: Distance by which to translate along the x-axis.
            dy: Distance by which to translate along the y-axis.
            dz: Distance by which to translate layer along the z-axis.
            inplace: If True, modifies the device (``self``) in-place and returns None,
                otherwise, creates a new device, translates it, and returns it.

        Returns:
            The translated device.
        """
        if inplace:
            device = self
        else:
            self._warn_if_mesh_exist("translate(..., inplace=False)")
            device = self.copy(with_mesh=False)
        for polygon in device.polygons:
            polygon.translate(dx, dy, inplace=True)
        if self.probe_points is not None:
            device.probe_points = self.probe_points + np.array([[dx, dy]])
        if device.mesh is not None:
            points = device.points
            points += np.array([[dx, dy]])
            device._create_dimensionless_mesh(points, device.triangles)
        if dz:
            device.layer.z0 += dz
        return device

    @contextmanager
    def translation(self, dx: float, dy: float, dz: float = 0) -> None:
        """A context manager that temporarily translates a device in-place,
        then returns it to its original position.

        Args:
            dx: Distance by which to translate polygons along the x-axis.
            dy: Distance by which to translate polygons along the y-axis.
            dz: Distance by which to translate layers along the z-axis.
        """
        try:
            self.translate(dx, dy, dz=dz, inplace=True)
            yield
        finally:
            self.translate(-dx, -dy, dz=-dz, inplace=True)

    def make_mesh(
        self,
        max_edge_length: Union[float, None] = None,
        min_points: Union[float, None] = None,
        smooth: int = 0,
        **meshpy_kwargs,
    ) -> None:
        """Generates and optimizes the triangular mesh.

        Args:
            min_points: Minimum number of vertices in the mesh. If None, then
                the number of vertices will be determined by meshpy_kwargs, the
                number of vertices in the underlying polygons, and ``max_edge_length``.
            max_edge_length: The maximum distance between vertices in the mesh.
                Passing a value <= 0 means that the number of mesh points will be
                determined solely by the density of points in the Device's film
                and abstract regions. Defaults to 1.5 * self.coherence_length.
            smooth: Number of Laplacian smoothing iterations to perform.
            **meshpy_kwargs: Passed to meshpy.triangle.build().
        """
        logger.info("Generating mesh...")
        boundary = self.film.points
        if max_edge_length is None:
            max_edge_length = 1.5 * self.coherence_length.magnitude
        points, triangles = generate_mesh(
            self.poly_points,
            hole_coords=[hole.points for hole in self.holes],
            min_points=min_points,
            max_edge_length=max_edge_length,
            boundary=boundary,
            **meshpy_kwargs,
        )
        if smooth:
            mesh = Mesh.from_triangulation(
                points, triangles, create_submesh=False
            ).smooth(smooth, create_submesh=False)
            points = mesh.sites
            triangles = mesh.elements
        logger.info(
            f"Finished generating mesh with {len(points)} points and "
            f"{len(triangles)} triangles."
        )
        self._create_dimensionless_mesh(points, triangles)

    def _create_dimensionless_mesh(
        self, points: np.ndarray, triangles: np.ndarray
    ) -> Mesh:
        """Creates the dimensionless mesh.

        Args:
            points: Mesh vertices in ``length_units``.
            triangles: Mesh triangle indices.

        Returns:
            The dimensionless ``Mesh`` object.
        """
        self.mesh = Mesh.from_triangulation(
            points / self.coherence_length.magnitude,
            triangles,
            create_submesh=True,
        )

    def mesh_stats_dict(self) -> Dict[str, Union[int, float, str]]:
        """Returns a dictionary of information about the mesh."""
        edge_lengths = self.edge_lengths
        areas = self.areas

        def _min(arr):
            if arr is not None:
                return arr.min()

        def _max(arr):
            if arr is not None:
                return arr.max()

        def _mean(arr):
            if arr is not None:
                return arr.mean()

        return dict(
            num_sites=len(self.mesh.sites) if self.mesh else None,
            num_elements=len(self.mesh.elements) if self.mesh else None,
            min_edge_length=_min(edge_lengths),
            max_edge_length=_max(edge_lengths),
            mean_edge_length=_mean(edge_lengths),
            min_area=_min(areas),
            max_area=_max(areas),
            mean_area=_mean(areas),
            coherence_length=self.coherence_length.magnitude,
            length_units=self.length_units,
        )

    def mesh_stats(self, precision: int = 3) -> HTML:
        """When called with in Jupyter notebook, displays
        a table of information about the mesh.

        Args:
            precision: Number of digits after the decimal for float values.

        Returns:
            An HTML table of mesh statistics.
        """
        stats = self.mesh_stats_dict()
        html = ["<table>", "<tr><b>Mesh Statistics</b></tr>"]
        for key, value in stats.items():
            if isinstance(value, float):
                value = f"{value:.{precision}e}"
            html.append(f"<tr><td><b>{key}</b></td><td>{value}</td></tr>")
        html.append("</table>")
        return HTML("".join(html))

    def plot(
        self,
        ax: Union[plt.Axes, None] = None,
        legend: bool = True,
        figsize: Union[Tuple[float, float], None] = None,
        mesh: bool = False,
        mesh_kwargs: Dict[str, Any] = dict(color="k", lw=0.5),
        **kwargs,
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Plot all of the device's polygons.

        Args:
            ax: matplotlib axis on which to plot. If None, a new figure is created.
            subplots: If True, plots each layer on a different subplot.
            legend: Whether to add a legend.
            figsize: matplotlib figsize, only used if ax is None.
            mesh: If True, plot the mesh.
            mesh_kwargs: Keyword arguments passed to ``ax.triplot()``
                if ``mesh`` is True.
            kwargs: Passed to ``ax.plot()`` for the polygon boundaries.

        Returns:
            Matplotlib Figure and Axes
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
        else:
            fig = ax.get_figure()
        points = self.points
        if mesh:
            if self.mesh is None:
                raise RuntimeError(
                    "Mesh does not exist. Run device.make_mesh() to generate the mesh."
                )
            x = points[:, 0]
            y = points[:, 1]
            tri = self.triangles
            ax.triplot(x, y, tri, **mesh_kwargs)
        for polygon in self.polygons:
            ax = polygon.plot(ax=ax, **kwargs)
        if self.probe_points is not None:
            ax.plot(*self.probe_points.T, "ko", label="Probe points")
        if legend:
            ax.legend(bbox_to_anchor=(1, 1), loc="upper left")
        units = ureg(self.length_units).units
        ax.set_xlabel(f"$x$ $[{units:~L}]$")
        ax.set_ylabel(f"$y$ $[{units:~L}]$")
        ax.set_aspect("equal")
        return fig, ax

    def patches(self) -> Dict[str, PathPatch]:
        """Returns a dict of ``{polygon_name: PathPatch}``
        for visualizing the device.
        """
        abstract_regions = self.abstract_regions
        hole_names = {hole.name for hole in self.holes}
        patches = dict()
        for polygon in self.polygons:
            if polygon.name in hole_names:
                continue
            coords = polygon.points.tolist()
            codes = [Path.LINETO for _ in coords]
            codes[0] = Path.MOVETO
            codes[-1] = Path.CLOSEPOLY
            poly = polygon.polygon
            for hole in self.holes:
                if polygon.name not in abstract_regions and poly.contains(hole.polygon):
                    hole_coords = hole.points.tolist()[::-1]
                    hole_codes = [Path.LINETO for _ in hole_coords]
                    hole_codes[0] = Path.MOVETO
                    hole_codes[-1] = Path.CLOSEPOLY
                    coords.extend(hole_coords)
                    codes.extend(hole_codes)
            patches[polygon.name] = PathPatch(Path(coords, codes))
        return patches

    def draw(
        self,
        ax: Union[plt.Axes, None] = None,
        legend: bool = True,
        figsize: Union[Tuple[float, float], None] = None,
        alpha: float = 0.5,
        exclude: Union[Union[str, List[str]], None] = None,
    ) -> Tuple[plt.Figure, Union[plt.Axes, np.ndarray]]:
        """Draws all polygons in the device as matplotlib patches.

        Args:
            ax: matplotlib axis on which to plot. If None, a new figure is created.
            legend: Whether to add a legend.
            figsize: matplotlib figsize, only used if ax is None.
            alpha: The alpha (opacity) value for the patches (0 <= alpha <= 1).
            exclude: A polygon name or list of polygon names to exclude
                from the figure.

        Returns:
            Matplotlib Figre and Axes.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
        else:
            fig = ax.get_figure()
        exclude = exclude or []
        if isinstance(exclude, str):
            exclude = [exclude]
        patches = self.patches()
        units = ureg(self.length_units).units
        x, y = self.poly_points.T
        margin = 0.1
        dx = np.ptp(x)
        dy = np.ptp(y)
        x0 = x.min() + dx / 2
        y0 = y.min() + dy / 2
        dx *= 1 + margin
        dy *= 1 + margin
        labels = []
        handles = []
        ax.set_aspect("equal")
        ax.grid(False)
        ax.set_xlim(x0 - dx / 2, x0 + dx / 2)
        ax.set_ylim(y0 - dy / 2, y0 + dy / 2)
        ax.set_xlabel(f"$x$ $[{units:~L}]$")
        ax.set_ylabel(f"$y$ $[{units:~L}]$")
        for i, (name, patch) in enumerate(patches.items()):
            if name in exclude:
                continue
            patch.set_alpha(alpha)
            patch.set_color(f"C{i % 10}")
            ax.add_artist(patch)
            labels.append(name)
            handles.append(patch)
        if self.probe_points is not None:
            (line,) = ax.plot(*self.probe_points.T, "ko", label="Probe points")
            handles.append(line)
            labels.append("Probe points")
        if legend:
            ax.legend(handles, labels, bbox_to_anchor=(1, 1), loc="upper left")
        return fig, ax

    def to_hdf5(
        self,
        path_or_group: Union[str, h5py.File, h5py.Group],
        save_mesh: bool = True,
    ) -> None:
        """Serializes the Device to disk.

        Args:
            path_or_group: A path to an HDF5 file, or an open HDF5 file or group
                into which to save the ``Device``.
            save_mesh: Whether to serialize the full mesh.
        """
        if isinstance(path_or_group, str):
            path = path_or_group
            if not path.endswith(".h5"):
                path = path + ".h5"
            if os.path.exists(path):
                raise IOError(f"Path already exists: {path}.")
            os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)
            save_context = h5py.File(path, "w-", libver="latest")
        else:
            h5_group = path_or_group
            save_context = nullcontext(h5_group)
        with save_context as f:
            f.attrs["name"] = self.name
            f.attrs["length_units"] = self.length_units
            self.layer.to_hdf5(f.create_group("layer"))
            self.film.to_hdf5(f.create_group("film"))
            for terminal in self.terminals:
                terminals_grp = f.require_group("terminals")
                terminal.to_hdf5(terminals_grp.create_group(terminal.name))
            if self.probe_points is not None:
                f["probe_points"] = self.probe_points
            for hole in sorted(self.holes, key=attrgetter("name")):
                group = f.require_group("holes")
                hole.to_hdf5(group.create_group(hole.name))
            for i, polygon in enumerate(self.abstract_regions):
                group = f.require_group("abstract_regions")
                polygon.to_hdf5(group.create_group(str(i)))
            if save_mesh and self.mesh is not None:
                self.mesh.to_hdf5(f.create_group("mesh"))

    @classmethod
    def from_hdf5(cls, path_or_group: Union[str, h5py.File, h5py.Group]) -> "Device":
        """Creates a new Device from one serialized to disk.

        Args:
            path_or_group: A path to an HDF5 file, or an open HDF5 file or group
                containing the serialized Device.

        Returns:
            The loaded Device instance.
        """
        if isinstance(path_or_group, str):
            h5_context = h5py.File(path_or_group, "r", libver="latest")
        else:
            if not isinstance(path_or_group, (h5py.File, h5py.Group)):
                raise TypeError(
                    f"Expected an h5py.File or h5py.Group, but got "
                    f"{type(path_or_group)}."
                )
            h5_context = nullcontext(path_or_group)
        terminals = probe_points = None
        holes = abstract_regions = mesh = None
        with h5_context as f:
            name = f.attrs["name"]
            length_units = f.attrs["length_units"]
            layer = Layer.from_hdf5(f["layer"])
            film = Polygon.from_hdf5(f["film"])
            if "terminals" in f:
                terminals = []
                for grp in f["terminals"].values():
                    terminals.append(Polygon.from_hdf5(grp))
            if "holes" in f:
                holes = [
                    Polygon.from_hdf5(grp)
                    for _, grp in sorted(f["holes"].items(), key=itemgetter(0))
                ]
            if "abstract_regions" in f:
                abstract_regions = [
                    Polygon.from_hdf5(grp)
                    for _, grp in sorted(
                        f["abstract_regions"].items(), key=itemgetter(0)
                    )
                ]
            if "probe_points" in f:
                probe_points = np.array(f["probe_points"])
            if "mesh" in f:
                mesh = Mesh.from_hdf5(f["mesh"])

        device = Device(
            name,
            layer=layer,
            film=film,
            holes=holes,
            abstract_regions=abstract_regions,
            terminals=terminals,
            probe_points=probe_points,
            length_units=length_units,
        )

        if mesh is not None:
            device.mesh = mesh

        return device

    def __repr__(self) -> str:
        # Normal tab "\t" renders a bit too big in jupyter if you ask me.
        indent = 4
        t = " " * indent
        nt = "\n" + t

        args = [
            f"{self.name!r}",
            f"layer={self.layer!r}",
            f"film={self.film!r}",
            f"holes={self.holes!r}",
            f"abstract_regions={self.abstract_regions!r}",
            f"terminals={self.terminals!r}",
            f"probe_points={self.probe_points!r}",
            f"length_units={self.length_units!r}",
        ]

        return f"{self.__class__.__name__}(" + nt + (", " + nt).join(args) + ",\n)"

    def __eq__(self, other) -> bool:
        if other is self:
            return True

        if not isinstance(other, Device):
            return False

        def compare(seq1, seq2, key="name"):
            key = attrgetter(key)
            return sorted(seq1, key=key) == sorted(seq2, key=key)

        if self.probe_points is None and other.probe_points is None:
            same_probe_points = True
        elif (
            isinstance(self.probe_points, np.ndarray)
            and isinstance(other.probe_points, np.ndarray)
            and np.allclose(self.probe_points, other.probe_points)
        ):
            same_probe_points = True
        else:
            same_probe_points = False

        return (
            self.name == other.name
            and self.layer == other.layer
            and self.film == other.film
            and compare(self.holes, other.holes)
            and compare(self.abstract_regions, other.abstract_regions)
            and compare(self.terminals, other.terminals)
            and same_probe_points
            and self.length_units == other.length_units
        )



import logging
from typing import Iterable, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import path
from scipy import interpolate
from shapely import affinity
from shapely import geometry as geo
from shapely.validation import explain_validity

from ..finite_volume.mesh import Mesh
from ..geometry import close_curve, ensure_unique
from .meshing import generate_mesh

logger = logging.getLogger(__name__)


PolygonType = Union[
    "Polygon",
    np.ndarray,
    geo.linestring.LineString,
    geo.polygon.LinearRing,
    geo.polygon.Polygon,
]


class Polygon:
    """A polygonal region located in a :class:`tdgl.Layer`.

    Args:
        name: Name of the polygon.
        points: The polygon vertices. This can be a shape ``(n, 2)`` array of x, y
            coordinates or a shapely ``LineString``, ``LinearRing``, or ``Polygon``.
        mesh: Whether to include this polygon when computing a mesh.
    """

    def __init__(
        self,
        name: Union[str, None] = None,
        *,
        points: PolygonType,
        mesh: bool = True,
    ):
        self.name = name
        self.points = points
        self.mesh = mesh

    @property
    def points(self) -> np.ndarray:
        """A shape ``(n, 2)`` array of counter-clockwise-oriented polygon vertices."""
        return self._points

    @points.setter
    def points(self, points) -> None:
        geom_types = (
            geo.linestring.LineString,
            geo.polygon.LinearRing,
            geo.polygon.Polygon,
        )
        if isinstance(points, Polygon):
            points = points.points
        if not isinstance(points, geom_types):
            points = np.asarray(points)
        points = geo.polygon.Polygon(points)
        points = geo.polygon.orient(points)
        if points.interiors:
            raise ValueError("Expected a simply-connected polygon.")
        if not points.is_valid:
            reason = explain_validity(points)
            raise ValueError(
                "The given points do not define a valid polygon for the following "
                f"reason: {reason}."
            )
        points = close_curve(np.array(points.exterior.coords))
        if points.ndim != 2 or points.shape[-1] != 2:
            raise ValueError(f"Expected shape (n, 2), but got {points.shape}.")
        self._points = points

    @property
    def is_valid(self) -> bool:
        """True if the ``Polygon`` has a ``name`` and its geometry is valid."""
        polygon = self.polygon
        return self.name is not None and polygon.is_valid and not polygon.interiors

    @property
    def area(self) -> float:
        """The area of the polygon."""
        return self.polygon.area

    @property
    def bbox(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Returns the coordinates of the lower left and upper right corners
        of the polygon's bounding box.
        """
        minx, miny, maxx, maxy = self.polygon.bounds
        return (minx, miny), (maxx, maxy)

    @property
    def extents(self) -> Tuple[float, float]:
        """Returns the total x, y extent of the polygon, (Delta_x, Delta_y)."""
        minx, miny, maxx, maxy = self.polygon.bounds
        return (maxx - minx), (maxy - miny)

    @property
    def polygon(self) -> geo.polygon.Polygon:
        """A shapely ``Polygon`` representing the Polygon."""
        return geo.polygon.Polygon(self.points)

    @property
    def path(self) -> path.Path:
        """A matplotlib.path.Path representing the polygon boundary."""
        return path.Path(self.points, closed=True)

    def contains_points(
        self,
        points: np.ndarray,
        index: bool = False,
        radius: float = 0,
    ) -> Union[bool, np.ndarray]:
        """Determines whether ``points`` lie within the polygon.

        Args:
            points: Shape ``(n, 2)`` array of x, y coordinates.
            index: If True, then return the indices of the points in ``points``
                that lie within the polygon. Otherwise, returns a shape ``(n, )``
                boolean array.
            radius: An additional margin on ``self.path``.
                See :meth:`matplotlib.path.Path.contains_points`.

        Returns:
            If index is True, returns the indices of the points in ``points``
            that lie within the polygon. Otherwise, returns a shape ``(n, )``
            boolean array indicating whether each point lies within the polygon.
        """
        bool_array = self.path.contains_points(np.atleast_2d(points), radius=radius)
        if index:
            return np.where(bool_array)[0]
        return bool_array

    def on_boundary(
        self, points: np.ndarray, radius: float = 1e-3, index: bool = False
    ):
        """Determines whether ``points`` lie within a given radius of the Polygon
        boundary.

        Args:
            points: Shape ``(n, 2)`` array of x, y coordinates.
            radius: Points within ``radius`` of the boundary are considered
                to lie on the boundary.
            index: If True, then return the indices of the points in ``points``
                that lie on the boundary. Otherwise, returns a shape ``(n, )``
                boolean array.

        Returns:
            If index is True, returns the indices of the points in ``points``
            that lie within the polygon. Otherwise, returns a shape ``(n, )``
            boolean array indicating whether each point lies within the polygon.
        """
        points = np.atleast_2d(points)
        p = self.path
        in_outer = p.contains_points(points, radius=radius)
        in_inner = p.contains_points(points, radius=-radius)
        boundary = np.logical_and(in_outer, ~in_inner)
        if index:
            return np.where(boundary)[0]
        return boundary

    def make_mesh(
        self,
        min_points: Union[int, None] = None,
        smooth: int = 0,
        **meshpy_kwargs,
    ) -> Mesh:
        """Returns the vertices and triangles of a Delaunay mesh covering the Polygon.

        Args:
            min_points: Minimum number of vertices in the mesh. If None, then
                the number of vertices will be determined by meshpy_kwargs and the
                number of vertices in the underlying polygons.
            smooth: Number of Laplacian smoothing steps to perform.
            **meshpy_kwargs: Passed to meshpy.triangle.build().

        Returns:
            Mesh vertex coordinates and triangle indices
        """
        points, triangles = generate_mesh(
            self.points,
            min_points=min_points,
            convex_hull=False,
            **meshpy_kwargs,
        )
        if smooth:
            mesh = Mesh.from_triangulation(
                points, triangles, create_submesh=False
            ).smooth(smooth)
        else:
            mesh = Mesh.from_triangulation(points, triangles)
        logger.debug(
            f"Finished generating mesh with {len(mesh.sites)} points and "
            f"{len(mesh.elements)} triangles."
        )
        return mesh

    def rotate(
        self,
        degrees: float,
        origin: Union[str, Tuple[float, float]] = (0.0, 0.0),
        inplace: bool = False,
    ) -> "Polygon":
        """Rotates the polygon counterclockwise by a given angle.

        Args:
            degrees: The amount by which to rotate the polygon.
            origin: (x, y) coorindates about which to rotate, or the strings
                "center" (for the bounding box center) or "centroid"
                (for the polygon center of mass).
            inplace: If True, modify the polygon in place. Otherwise, return
                a modified copy.

        Returns:
            The rotated polygon.
        """
        polygon = self if inplace else self.copy()
        polygon.points = affinity.rotate(
            self.polygon, degrees, origin=origin, use_radians=False
        )
        return polygon

    def translate(
        self,
        dx: float = 0.0,
        dy: float = 0.0,
        inplace: bool = False,
    ) -> "Polygon":
        """Translates the polygon by a given distance.

        Args:
            dx: Distance by which to translate along the x-axis.
            dy: Distance by which to translate along the y-axis.
            inplace: If True, modify the polygon in place. Otherwise, return
                a modified copy.

        Returns:
            The translated polygon.
        """
        polygon = self if inplace else self.copy()
        polygon.points = affinity.translate(self.polygon, xoff=dx, yoff=dy)
        return polygon

    def scale(
        self,
        xfact: float = 1.0,
        yfact: float = 1.0,
        origin: Union[str, Tuple[float, float]] = (0, 0),
        inplace: bool = False,
    ) -> "Polygon":
        """Scales the polygon horizontally by ``xfact`` and vertically by ``yfact``.

        Negative ``xfact`` (``yfact``) can be used to reflect the polygon horizontally
        (vertically) about the ``origin``.

        Args:
            xfact: Distance by which to translate along the x-axis.
            yfact: Distance by which to translate along the y-axis.
            origin: (x, y) coorindates for the scaling origin, or the strings
                "center" (for the bounding box center) or "centroid"
                (for the polygon center of mass).
            inplace: If True, modify the polygon in place. Otherwise, return
                a modified copy.

        Returns:
            The scaled polygon.
        """
        polygon = self if inplace else self.copy()
        polygon.points = affinity.scale(
            self.polygon, xfact=xfact, yfact=yfact, origin=origin
        )
        return polygon

    def _join_via(
        self,
        other: PolygonType,
        operation: str,
    ) -> geo.polygon.Polygon:
        """Joins ``self.polygon`` with another polygon-like object
        via a given operation.
        """
        valid_types = (
            np.ndarray,
            geo.linestring.LineString,
            geo.polygon.LinearRing,
            geo.polygon.Polygon,
        )
        valid_operations = (
            "union",
            "intersection",
            "difference",
        )
        if operation not in valid_operations:
            raise ValueError(
                f"Unknown operation: {operation}. "
                f"Valid operations are {valid_operations}."
            )
        if isinstance(other, Polygon):
            other_poly = other.polygon
        elif isinstance(other, valid_types):
            other_poly = geo.polygon.Polygon(other)
        if not isinstance(other_poly, geo.polygon.Polygon):
            raise TypeError(
                f"Valid types are {(Polygon, ) + valid_types}, got {type(other)}."
            )
        joined = getattr(self.polygon, operation)(other_poly)
        reason = None
        if not isinstance(joined, geo.polygon.Polygon):
            reason = f"joined polygon has an unexpected type ({type(joined)})"
        elif joined.is_empty:
            reason = "joined polygon is empty"
        elif not joined.is_valid:
            reason = explain_validity(joined)
        if reason is not None:
            raise ValueError(
                f"The {operation} of the two polygons is not a valid polygon "
                f"for the following reason: {reason}."
            )
        return joined

    def union(
        self,
        *others: PolygonType,
        name: Union[str, None] = None,
    ) -> "Polygon":
        """Returns the union of the polygon and zero or more other polygons.

        Args:
            others: One or more objects with which to join the polygon.
            name: A name for the resulting joined Polygon (defaults to ``self.name``.)

        Returns:
            A new :class:`Polygon` instance representing the union
            of ``self`` and ``others``.
        """
        if not others:
            return self.copy()
        first, *rest = others
        return Polygon(
            name=name or self.name,
            points=self._join_via(first, "union"),
            mesh=self.mesh,
        ).union(*rest, name=name)

    def intersection(
        self,
        *others: PolygonType,
        name: Union[str, None] = None,
    ) -> "Polygon":
        """Returns the intersection of the polygon and zero or more other polygons.

        Args:
            others: One or more objects with which to join the polygon.
            name: A name for the resulting joined Polygon (defaults to ``self.name``.)

        Returns:
            A new :class:`Polygon` instance representing the intersection
            of ``self`` and ``others``.
        """
        if not others:
            return self.copy()
        first, *rest = others
        return Polygon(
            name=name or self.name,
            points=self._join_via(first, "intersection"),
            mesh=self.mesh,
        ).intersection(*rest, name=name)

    def difference(
        self,
        *others: PolygonType,
        name: Union[str, None] = None,
    ) -> "Polygon":
        """Returns the difference of the polygon and zero more other polygons.

        Args:
            others: One or more objects with which to join the polygon.
            name: A name for the resulting joined Polygon (defaults to ``self.name``.)

        Returns:
            A new :class:`Polygon` instance representing the difference
            of ``self`` and ``others``.

        .. _shapely documentation: https://shapely.readthedocs.io/en/stable/manual.html
        """
        if not others:
            return self.copy()
        first, *rest = others
        return Polygon(
            name=name or self.name,
            points=self._join_via(first, "difference"),
            mesh=self.mesh,
        ).difference(*rest, name=name)

    def __add__(self, other: PolygonType) -> "Polygon":
        return self.union(other)

    def __sub__(self, other: PolygonType) -> "Polygon":
        return self.difference(other)

    def __mul__(self, other: PolygonType) -> "Polygon":
        return self.intersection(other)

    def buffer(
        self,
        distance: float,
        join_style: Union[str, int] = "mitre",
        mitre_limit: float = 5.0,
        single_sided: bool = True,
        as_polygon: bool = True,
    ) -> Union[np.ndarray, "Polygon"]:
        """Returns polygon points or a new Polygon object with vertices offset from
        ``self.points`` by a given ``distance``. If ``distance > 0`` this "inflates"
        the polygon, and if ``distance < 0`` this shrinks the polygon.


        Args:
            distance: The amount by which to inflate or deflate the polygon.
            join_style: One of "round" (1), "mitre" (2), or "bevel" (3).
                See the `shapely documentation`_.
            mitre_limit: See the `shapely documentation`_.
            single_sided: See the `shapely documentation`_.
            as_polygon: If True, returns a new ``Polygon`` instance, otherwise
                returns a shape ``(n, 2)`` array of polygon vertices.

        Returns:
            A new ``Polygon`` or an array of vertices offset by ``distance``.

        .. _shapely documentation: https://shapely.readthedocs.io/en/stable/manual.html
        """
        if isinstance(join_style, str):
            join_style = getattr(geo.JOIN_STYLE, join_style)
        poly = self.polygon.buffer(
            distance,
            join_style=join_style,
            mitre_limit=mitre_limit,
            single_sided=single_sided,
        )

        polygon = Polygon(
            name=self.name,
            points=poly,
            mesh=self.mesh,
        )
        npts = max(len(polygon.points), len(self.points))
        polygon = polygon.resample(npts)
        if as_polygon:
            return polygon
        return polygon.points

    def resample(
        self, num_points: Union[int, None] = None, degree: int = 1, smooth: float = 0
    ) -> "Polygon":
        """Resample vertices so that they are approximately uniformly distributed
        along the polygon boundary.

        Args:
            num_points: Number of points to interpolate to. If ``num_points`` is None,
                the polygon is resampled to ``len(self.points)`` points. If
                ``num_points`` is not None and has a boolean value of False,
                then an unaltered copy of the polygon is returned.
            degree: The degree of the spline with which to iterpolate.
                Defaults to 1 (linear spline).
            smooth: Smoothing condition.

        """
        if num_points is None:
            num_points = len(self.points)
        if not num_points:
            return self.copy()
        points = ensure_unique(self.points.copy())
        tck, _ = interpolate.splprep(points.T, k=degree, s=smooth)
        x, y = interpolate.splev(np.linspace(0, 1, num_points - 1), tck)
        points = close_curve(np.array([x, y]).T)
        return Polygon(
            name=self.name,
            points=points,
            mesh=self.mesh,
        )

    def set_name(self, name: Union[str, None]) -> "Polygon":
        """Sets the Polygon's name and returns ``self``."""
        self.name = name
        return self

    def plot(self, ax: Union[plt.Axes, None] = None, **kwargs) -> plt.Axes:
        """Plots the Polygon's vertices.

        Args:
            ax: The matplotlib Axes on which to plot. If None is given, a new one
                is created.
            kwargs: Passed to ``ax.plot()``.

        Returns:
            The matplotlib Axes.
        """
        if ax is None:
            _, ax = plt.subplots()
        kwargs = kwargs.copy()
        kwargs["label"] = self.name
        ax.plot(*self.points.T, **kwargs)
        ax.set_aspect("equal")
        return ax

    @classmethod
    def from_union(
        cls,
        items: Iterable[PolygonType],
        *,
        name: Union[str, None] = None,
        mesh: bool = True,
    ) -> "Polygon":
        """Creates a new :class:`Polygon` from the union of a sequence of polygons.

        Args:
            items: A sequence of polygon-like objects to join.
            name: Name of the polygon.
            mesh: Whether to include this polygon when computing a mesh.

        Returns:
            A new :class:`Polygon`.
        """
        first, *rest = items
        polygon = cls(name=name, points=first, mesh=mesh)
        return polygon.union(*rest)

    @classmethod
    def from_intersection(
        cls,
        items: Iterable[PolygonType],
        *,
        name: Union[str, None] = None,
        mesh: bool = True,
    ) -> "Polygon":
        """Creates a new :class:`Polygon` from the intersection
        of a sequence of polygons.

        Args:
            items: A sequence of polygon-like objects to join.
            name: Name of the polygon.
            mesh: Whether to include this polygon when computing a mesh.

        Returns:
            A new :class:`Polygon`.
        """
        first, *rest = items
        polygon = cls(name=name, points=first, mesh=mesh)
        return polygon.intersection(*rest)

    @classmethod
    def from_difference(
        cls,
        items: Iterable[PolygonType],
        *,
        name: Union[str, None] = None,
        mesh: bool = True,
    ) -> "Polygon":
        """Creates a new :class:`Polygon` from the difference
        of a sequence of polygons.

        Args:
            items: A sequence of polygon-like objects to join.
            name: Name of the polygon.
            mesh: Whether to include this polygon when computing a mesh.

        Returns:
            A new :class:`Polygon`.
        """
        first, *rest = items
        polygon = cls(name=name, points=first, mesh=mesh)
        return polygon.difference(*rest)

    def to_hdf5(self, h5_group: h5py.Group) -> None:
        """Save the ``Polygon`` to an :class:`h5py.Group`."""
        if self.name is not None:
            h5_group.attrs["name"] = self.name
        h5_group.attrs["mesh"] = self.mesh
        h5_group["points"] = self.points

    @classmethod
    def from_hdf5(cls, h5_group: h5py.Group) -> "Polygon":
        """Load a ``Polygon`` from an :class:`h5py.Group`."""
        name = None
        if "name" in h5_group.attrs:
            name = h5_group.attrs["name"]
        return Polygon(
            name=name,
            points=np.array(h5_group["points"]),
            mesh=h5_group.attrs["mesh"],
        )

    def __repr__(self) -> str:
        name = f'"{self.name}"' if self.name is not None else None
        return (
            f"{self.__class__.__name__}(name={name}, "
            f"points=<ndarray: shape={self.points.shape}>, mesh={self.mesh})"
        )

    def __eq__(self, other) -> bool:
        if other is self:
            return True

        if not isinstance(other, Polygon):
            return False

        return self.name == other.name and np.allclose(self.points, other.points)

    def copy(self) -> "Polygon":
        """Returns a deep copy of the :class:`Polygon`"""
        return Polygon(
            self.name,
            points=self.points.copy(),
            mesh=self.mesh,
        )



from typing import Union

import h5py


class Layer:
    """A superconducting thin film.

    Args:
        london_lambda: The London penetration depth of the film.
        coherence_length: The superconducting coherence length of the film.
        thickness: The thickness of the film.
        conductivity: The normal state conductivity of the superconductor in
            Siemens / length_unit.
        u: The ratio of the relaxation times for the order parameter amplitude
            and phase. This value is 5.79 for dirty superconductors.
        gamma: This parameter quantifies the effect of inelastic phonon-electron
            scattering. :math:`\\gamma` is proportional to the inelastic scattering
            time and the size of the superconducting gap.
        z0: Vertical location of the film.
    """

    def __init__(
        self,
        *,
        london_lambda: float,
        coherence_length: float,
        thickness: float,
        conductivity: Union[float, None] = None,
        u: float = 5.79,
        gamma: float = 10.0,
        z0: float = 0,
    ):
        self.london_lambda = london_lambda
        self.coherence_length = coherence_length
        self.thickness = thickness
        self.conductivity = conductivity
        self.u = u
        self.gamma = gamma
        self.z0 = z0

    @property
    def Lambda(self) -> float:
        """Effective magnetic penetration depth, :math:`\\Lambda=\\lambda^2/d`."""
        return self.london_lambda**2 / self.thickness

    def copy(self) -> "Layer":
        """Create a deep copy of the :class:`tdgl.Layer`."""
        return Layer(
            london_lambda=self.london_lambda,
            coherence_length=self.coherence_length,
            thickness=self.thickness,
            conductivity=self.conductivity,
            u=self.u,
            gamma=self.gamma,
            z0=self.z0,
        )

    def to_hdf5(self, h5_group: h5py.Group) -> None:
        """Save the :class:`tdgl.Layer` to an :class:`h5py.Group`.

        Args:
            h5_group: An open :class:`h5py.Group` to which to save the layer.
        """
        h5_group.attrs["london_lambda"] = self.london_lambda
        h5_group.attrs["coherence_length"] = self.coherence_length
        h5_group.attrs["thickness"] = self.thickness
        h5_group.attrs["u"] = self.u
        h5_group.attrs["gamma"] = self.gamma
        h5_group.attrs["z0"] = self.z0
        if self.conductivity is not None:
            h5_group.attrs["conductivity"] = self.conductivity

    @staticmethod
    def from_hdf5(h5_group: h5py.Group) -> "Layer":
        """Load a :class:`tdgl.Layer` from an :class:`h5py.Group`.

        Args:
            h5_group: An open :class:`h5py.Group` from which to load the layer.

        Returns:
            A new :class:`tdgl.Layer` instance.
        """

        def get(key, default=None):
            if key in h5_group.attrs:
                return h5_group.attrs[key]
            return default

        return Layer(
            london_lambda=get("london_lambda"),
            coherence_length=get("coherence_length"),
            thickness=get("thickness"),
            conductivity=get("conductivity"),
            u=get("u"),
            gamma=get("gamma"),
            z0=get("z0"),
        )

    def __eq__(self, other):
        if self is other:
            return True

        if not isinstance(other, Layer):
            return False

        return (
            self.london_lambda == other.london_lambda
            and self.coherence_length == other.coherence_length
            and self.thickness == other.thickness
            and self.conductivity == other.conductivity
            and self.u == other.u
            and self.gamma == other.gamma
            and self.z0 == other.z0
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"london_lambda={self.london_lambda}, "
            f"coherence_length={self.coherence_length}, "
            f"thickness={self.thickness}, "
            f"conducivitiy={self.conductivity}, "
            f"u={self.u}, "
            f"gamma={self.gamma}, "
            f"z0={self.z0}"
            f")"
        )



import dataclasses
import logging
import operator
import os
import pickle
import shutil
from contextlib import nullcontext
from datetime import datetime
from typing import Any, Callable, Dict, NamedTuple, Optional, Tuple, Union

import cloudpickle
import h5py
import matplotlib.pyplot as plt
import numpy as np
import pint
from scipy import interpolate
from scipy.spatial import distance

from ..about import version_dict
from ..device.device import Device
from ..device.polygon import Polygon
from ..em import biot_savart_2d, convert_field
from ..finite_volume.operators import build_gradient
from ..fluxoid import Fluxoid
from ..geometry import path_vectors
from ..parameter import Parameter
from ..solver.runner import SolverOptions
from .data import DynamicsData, TDGLData, get_data_range, get_edge_quantity_data

logger = logging.getLogger(__name__)


class BiotSavartField(NamedTuple):
    """The magnetic field due to a current distribution, with the field due to the
    supercurrent and normal current labeled separately.

    Args:
        supercurrent: An array of fields due to the supercurrent.
        normal_current: An array of fields due to the normal current.
    """

    supercurrent: np.ndarray
    normal_current: np.ndarray


class BoundaryPhases(NamedTuple):
    """A container for the phase of the order parameter along a polygon boundary.

    Args:
        indices: The mesh vertex indices of the boundary.
        phase: The phase of the order parameter at each vertex on the boundary.
    """

    indices: np.ndarray
    phases: np.ndarray


class Solution:
    """A container for the results of a TDGL simulation.

    Args:
        device: The :class:`tdgl.Device` that was solved
        options: A :class:`tdgl.SolverOptions` instance.
        path: Path to the HDF5 file containing the raw output data.
        applied_vector_potential: The ``Parameter`` defining the applied vector potential.
        terminal_currents: A dict of ``{terminal_name: current}`` or a callable with signature
            ``func(time) -> {terminal_name: current}``, where ``current`` is a float
            in units of ``current_units``.
        disorder_epsilon: The disorder parameter :math:`\\epsilon`. If
            :math:`\\epsilon(\\mathbf{r}) < 1` weakens the order parameter at position
            :math:`\\mathbf{r}`, which can be used to model inhomogeneity.
        total_seconds: The total wall time in seconds.
    """

    def __init__(
        self,
        *,
        device: Device,
        options: SolverOptions,
        path: os.PathLike,
        applied_vector_potential: Parameter,
        terminal_currents: Union[Dict[str, float], Callable],
        disorder_epsilon: Union[float, Callable],
        total_seconds: float,
        _solve_step: int = -1,
    ):
        self.device = device.copy()
        self.device.mesh = device.mesh
        self.options = options
        self.path = path
        self.applied_vector_potential = applied_vector_potential
        self.terminal_currents = terminal_currents
        self.disorder_epsilon = disorder_epsilon

        self.data_range: Union[Tuple[int, int], None] = None
        """A tuple of ``(min_step, max_step)``."""
        self.supercurrent_density: Union[np.ndarray, None] = None
        """Sheet supercurrent density, :math:`\\mathbf{K}_s`"""
        self.normal_current_density: Union[np.ndarray, None] = None
        """Sheet normal density, :math:`\\mathbf{K}_n`"""
        self._vorticity: Union[np.ndarray, None] = None

        # Make field_units and current_units "read-only" attributes.
        # The should never be changed after instantiation.
        self._field_units = str(self.options.field_units)
        self._current_units = str(self.options.current_units)
        self._time_created = datetime.now()
        self.total_seconds = total_seconds

        self.tdgl_data: Union[TDGLData, None] = None
        """A container for the raw TDGL data (in dimensionless units)."""
        self.dynamics: Union[DynamicsData, None] = None
        """A container for the time dynamics of the solution (in dimensionless units)."""
        self._solve_step = _solve_step
        self.load_tdgl_data(self._solve_step)
        self._version_info = version_dict()

    @property
    def saved_on_disk(self) -> bool:
        """Returns ``True`` if the underlying HDF5 file exists on disk."""
        return os.path.exists(self.path)

    @property
    def solve_step(self) -> int:
        """The solver iteration corresponding to the current
        :class:`tdgl.solution.data.TDGLData`.

        Setting ``solve_step`` automatically loads data for the specitied step.
        """
        return self._solve_step

    @solve_step.setter
    def solve_step(self, step: int) -> None:
        self.load_tdgl_data(solve_step=step)

    @property
    def times(self) -> Union[np.ndarray, None]:
        """The time associated with each solve step."""
        if self.dynamics is None:
            return None
        times = self.dynamics.time
        step = self.options.save_every
        saved_times = times[::step]
        if saved_times[-1] == times[-1]:
            return saved_times.copy()
        # Append the final time step in the simulation, which is always saved.
        return np.concatenate([saved_times, times[-1:]])

    def closest_solve_step(self, time: float) -> int:
        """Returns the index of the saved step closest in time to ``time``.

        Args:
            time: The time for which to find the closest index.

        Returns:
            The index of the saved solve step whose time is closest to ``time``
        """
        return np.argmin(np.abs(self.times - time))

    def load_tdgl_data(
        self, solve_step: int = -1, h5file: Union[h5py.File, None] = None
    ) -> None:
        """Loads the TDGL results from file for a given solve step.

        Args:
            solve_step: The step index for which to load data.
                Defaults to -1, i.e. the final step.
        """
        if h5file is None:
            read_context = h5py.File(self.path, "r", libver="latest")
        else:
            read_context = nullcontext(h5file)
        with read_context as f:
            self.data_range = step_min, step_max = get_data_range(f)
            if solve_step == 0:
                step = step_min
            elif solve_step < 0:
                step = step_max + 1 + solve_step
            else:
                step = solve_step
            self.tdgl_data = TDGLData.from_hdf5(f, step)
            self.dynamics = DynamicsData.from_hdf5(f, *self.data_range)
        mesh = self.device.mesh
        self._solve_step = step
        supercurrent, sc_direc, _ = get_edge_quantity_data(
            self.tdgl_data.supercurrent, mesh
        )
        normal_current, nc_direc, _ = get_edge_quantity_data(
            self.tdgl_data.normal_current, mesh
        )
        K0 = self.device.K0.to(f"{self.current_units} / {self.device.length_units}")
        # Current density, evaluated on the mesh edges.
        self.supercurrent_density = K0 * supercurrent[:, np.newaxis] * sc_direc
        self.normal_current_density = K0 * normal_current[:, np.newaxis] * nc_direc
        self._vorticity = None

    def _compute_vorticity(self) -> None:
        device = self.device
        mesh = device.mesh
        # Calculate the vorticity, evaluated on mesh sites.
        # The vorticity is the curl of the current density.
        j_sc_site = mesh.get_quantity_on_site(self.tdgl_data.supercurrent)
        j_nm_site = mesh.get_quantity_on_site(self.tdgl_data.normal_current)
        j_site = j_sc_site + j_nm_site
        gradient = build_gradient(mesh)
        normalized_directions = (
            mesh.edge_mesh.directions
            / np.linalg.norm(mesh.edge_mesh.directions, axis=1)[:, np.newaxis]
        )
        grad_jx = gradient @ j_site[:, 0]
        grad_jy = gradient @ j_site[:, 1]
        djy_dx = grad_jy * normalized_directions[:, 0]
        djx_dy = grad_jx * normalized_directions[:, 1]
        vorticity_on_edges = djy_dx - djx_dy
        vorticity = mesh.get_quantity_on_site(vorticity_on_edges, vector=False)
        scale = (device.K0 / device.coherence_length).to(
            f"{self.current_units} / {self.device.length_units}**2"
        )
        self._vorticity = vorticity * scale

    @property
    def vorticity(self) -> Union[np.ndarray, None]:
        """The current vorticity,
        :math:`\\omega=(\\nabla\\times\\mathbf{K})\\cdot\\hat{\\mathbf{z}}`
        """
        if self.supercurrent_density is None:
            return None
        if self._vorticity is None:
            self._compute_vorticity()
        return self._vorticity

    @property
    def current_density(self) -> np.ndarray:
        """The total sheet current density,
        :math:`\\mathbf{K}=\\mathbf{K}_s+\\mathbf{K}_n`.
        """
        if self.supercurrent_density is None:
            return None
        return self.supercurrent_density + self.normal_current_density

    @property
    def field_units(self) -> str:
        """The units in which magnetic fields are specified."""
        return self._field_units

    @property
    def current_units(self) -> str:
        """The units in which currents are specified."""
        return self._current_units

    @property
    def time_created(self) -> datetime:
        """The time at which the solution was originally created."""
        return self._time_created

    @property
    def version_info(self) -> Dict[str, str]:
        """A dictionary of dependency versions."""
        return self._version_info

    def grid_current_density(
        self,
        *,
        dataset: Union[str, None] = None,
        grid_shape: Union[int, Tuple[int, int]] = (200, 200),
        method: str = "linear",
        units: Union[str, None] = None,
        with_units: bool = False,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Interpolates the sheet current density to a rectangular grid.

        Keyword arguments are passed to :func:`scipy.interpolate.griddata`.

        .. seealso::

            :meth:`tdgl.Solution.interp_current_density`

        Args:
            dataset: The dataset to interpolate. One of ``"supercurrent"``,
                ``"normal_current"``, or ``None``. If ``None``, then the total
                sheet current density is used.
            grid_shape: Shape of the desired rectangular grid. If a single integer
                N is given, then the grid will be square, shape = (N, N).
            method: Interpolation method to use (see :func:`scipy.interpolate.griddata`).
            units: The desired units for the current density. Defaults to
                ``self.current_units / self.device.length_units``.
            with_units: Whether to return a :class:`pint.Quantity` array
                with units attached.

        Returns:
            x grid, y grid, interpolated current density
        """
        if dataset is None:
            J = self.current_density
        elif dataset == "supercurrent":
            J = self.supercurrent_density
        elif dataset == "normal_current":
            J = self.normal_current_density
        else:
            raise ValueError(f"Unexpected dataset: {dataset}.")

        units = units or f"{self.current_units} / {self.device.length_units}"
        J = J.to(units)
        if isinstance(grid_shape, int):
            grid_shape = (grid_shape, grid_shape)
        points = self.device.points
        x = points[:, 0]
        y = points[:, 1]
        xgrid, ygrid = np.meshgrid(
            np.linspace(x.min(), x.max(), grid_shape[1]),
            np.linspace(y.min(), y.max(), grid_shape[0]),
        )
        Jx = interpolate.griddata(
            points, J[:, 0].magnitude, (xgrid, ygrid), method=method, **kwargs
        ).ravel()
        Jy = interpolate.griddata(
            points, J[:, 1].magnitude, (xgrid, ygrid), method=method, **kwargs
        ).ravel()
        xy = np.array([xgrid.ravel(), ygrid.ravel()]).T
        hole_mask = np.logical_or.reduce(
            [hole.contains_points(xy) for hole in self.device.holes]
        )
        Jx[hole_mask] = 0
        Jy[hole_mask] = 0
        Jgrid = np.array([Jx.reshape(grid_shape), Jy.reshape(grid_shape)])
        if with_units:
            length_units = self.device.ureg(self.device.length_units)
            xgrid = xgrid * length_units
            ygrid = ygrid * length_units
            Jgrid = (Jgrid * J.units).to(units)
        return xgrid, ygrid, Jgrid

    def interp_current_density(
        self,
        positions: np.ndarray,
        *,
        dataset: Union[str, None] = None,
        method: str = "linear",
        units: Union[str, None] = None,
        with_units: bool = False,
    ) -> np.ndarray:
        """Interpolates the sheet current density at unstructured coordinates.

        .. seealso::

            :meth:`tdgl.Solution.grid_current_density`

        Args:
            positions: Shape ``(m, 2)`` array of x, y coordinates at which to evaluate
                the current density.
            dataset: The dataset to interpolate. One of ``"supercurrent"``,
                ``"normal_current"``, or ``None``. If ``None``, then the total
                sheet current density is used.
            method: Interpolation method to use, ``"nearest"``, ``"linear"``,
                or ``"cubic"``.
            units: The desired units for the current density. Defaults to
                ``self.current_units / self.device.length_units``.
            with_units: Whether to return a :class:`pint.Quantity` array
                with units attached.

        Returns:
            The interpolated current density as an array of floats
            or a :class:`pint.Quantity` array.
        """
        valid_methods = ("nearest", "linear", "cubic")
        if method not in valid_methods:
            raise ValueError(
                f"Interpolation method must be one of {valid_methods} (got {method})."
            )
        if method == "nearest":
            interpolator = interpolate.NearestNDInterpolator
            interp_kwargs = dict()
        elif method == "linear":
            interpolator = interpolate.LinearNDInterpolator
            interp_kwargs = dict(fill_value=0)
        else:  # "cubic"
            interpolator = interpolate.CloughTocher2DInterpolator
            interp_kwargs = dict(fill_value=0)

        if dataset is None:
            J = self.current_density
        elif dataset == "supercurrent":
            J = self.supercurrent_density
        elif dataset == "normal_current":
            J = self.normal_current_density
        else:
            raise ValueError(f"Unexpected dataset: {dataset}.")

        if units is None:
            units = f"{self.current_units} / {self.device.length_units}"
        positions = np.atleast_2d(positions)
        xy = self.device.points
        J_interp = interpolator(xy, J.to(units).magnitude, **interp_kwargs)
        J = J_interp(positions)
        J[~np.isfinite(J)] = 0
        J[~self.device.contains_points(positions)] = 0
        if with_units:
            J = J * self.device.ureg(units)
        return J

    def interp_order_parameter(
        self,
        positions: np.ndarray,
        method: str = "linear",
    ) -> np.ndarray:
        """Interpolates the order parameter at unstructured coordinates.

        Args:
            positions: Shape ``(m, 2)`` array of x, y coordinates at which to evaluate
                the order parameter.
            method: Interpolation method to use, ``"nearest"``, ``"linear"``,
                or ``"cubic"``.

        Returns:
            The interpolated order parameter.
        """
        valid_methods = ("nearest", "linear", "cubic")
        if method not in valid_methods:
            raise ValueError(
                f"Interpolation method must be one of {valid_methods} (got {method})."
            )
        if method == "nearest":
            interpolator = interpolate.NearestNDInterpolator
            interp_kwargs = dict()
        elif method == "linear":
            interpolator = interpolate.LinearNDInterpolator
            interp_kwargs = dict(fill_value=1)
        else:  # "cubic"
            interpolator = interpolate.CloughTocher2DInterpolator
            interp_kwargs = dict(fill_value=1)
        positions = np.atleast_2d(positions)
        xy = self.device.points
        psi = self.tdgl_data.psi
        psi_interp = interpolator(xy, psi, **interp_kwargs)
        return psi_interp(positions)

    def polygon_fluxoid(
        self,
        polygon_points: Union[np.ndarray, Polygon],
        interp_method: str = "linear",
        units: str = "Phi_0",
        with_units: bool = True,
    ) -> Fluxoid:
        """Computes the :class:`tdgl.Fluxoid` (flux + supercurrent) for
        a given polygonal region.

        The fluxoid for a closed region :math:`S` with boundary :math:`\\partial S`
        is defined as:

        .. math::

            \\begin{split}
            \\Phi^f_S &= \\Phi^f_{S,\\text{ flux}} + \\Phi^f_{S,\\text{ supercurrent}}
            \\\\&=\\int_S \\mu_0 H_z(\\mathbf{r})\\,\\mathrm{d}^2r +
                \\oint_{\\partial S}
                \\mu_0\\Lambda(\\mathbf{r})\\mathbf{K}_s(\\mathbf{r})\\cdot\\mathrm{d}\\mathbf{r}
            \\\\&=\\oint_{\\partial S} \\mathbf{A}(\\mathbf{r})\\cdot\\mathrm{d}\\mathbf{r} +
                \\oint_{\\partial S}
                \\mu_0\\Lambda(\\mathbf{r})\\mathbf{K}_s(\\mathbf{r})\\cdot\\mathrm{d}\\mathbf{r}
            \\end{split}

        .. seealso::

            :class:`tdgl.Fluxoid`, :func:`tdgl.make_fluxoid_polygons`

        Args:
            polygon_points: A shape ``(n, 2)`` array of ``(x, y)`` coordinates of
                polygon vertices defining the closed region :math:`S`.
            interp_method: Interpolation method to use, ``"nearest"``, ``"linear"``,
                or ``"cubic"``.
            units: The desired units for the fluxoid.
            with_units: Whether to return values as :class:`pint.Quantity` instances
                with units attached.

        Returns:
            The polygon's :class:`Fluxoid`.
        """
        device = self.device
        ureg = device.ureg
        if units is None:
            units = f"{self.field_units} * {self.device.length_units} ** 2"
        polygon = Polygon(points=polygon_points)
        points = polygon.points
        if not device.film.contains_points(points).all():
            raise ValueError(
                "The polygon must lie completely within the superconducting film."
            )
        # Evaluate the supercurrent density at the polygon coordinates.
        J_units = f"{self.current_units} / {device.length_units}"
        J_poly = self.interp_current_density(
            points,
            dataset="supercurrent",
            method=interp_method,
            units=J_units,
            with_units=True,
        )
        zs = device.layer.z0 * np.ones(len(points))
        dl = np.diff(points, axis=0, prepend=points[:1]) * ureg(device.length_units)
        A_units = f"{self.field_units} * {device.length_units}"
        A_poly = self.vector_potential_at_position(
            points,
            zs=zs,
            units=A_units,
            with_units=True,
            return_sum=True,
        )[:, :2]
        # Compute the flux part of the fluxoid:
        # \oint_{\\partial poly} \vec{A}\cdot\mathrm{d}\vec{r}
        int_A = np.trapz((A_poly * dl).sum(axis=1))
        flux_part = int_A.to(units)
        # Compute the supercurrent part of the fluxoid:
        # \oint_{poly}\Lambda\vec{J}\cdot\mathrm{d}\vec{r}
        Lambda = device.layer.Lambda
        psi_poly = self.interp_order_parameter(points, method=interp_method)
        ns = np.abs(psi_poly) ** 2
        Lambda = Lambda / ns * ureg(device.length_units)
        int_J = np.trapz((Lambda[:, np.newaxis] * J_poly * dl).sum(axis=1))
        supercurrent_part = (ureg("mu_0") * int_J).to(units)
        if not with_units:
            flux_part = flux_part.magnitude
            supercurrent_part = supercurrent_part.magnitude
        return Fluxoid(flux_part, supercurrent_part)

    def hole_fluxoid(
        self,
        hole_name: str,
        points: Union[np.ndarray, None] = None,
        interp_method: str = "linear",
        units: str = "Phi_0",
        with_units: bool = True,
    ) -> Fluxoid:
        """Calculcates the fluxoid for a polygon enclosing the specified hole.

        .. seealso::

            :meth:`tdgl.Solution.polygon_fluxoid`, :meth:`tdgl.Solution.boundary_phases`

        Args:
            hole_name: The name of the hole for which to calculate the fluxoid.
            points: The vertices of the polygon enclosing the hole. If None is given,
                a polygon is generated using
                :func:`tdgl.make_fluxoid_polygons`.
            interp_method: Interpolation method to use, ``"nearest"``, ``"linear"``,
                or ``"cubic"``.
            units: The desired units for the fluxoid.
            with_units: Whether to return values as :class:`pint.Quantity` instances
                with units attached.

        Returns:
            The hole's :class:`tdgl.Fluxoid`.
        """
        if points is None:
            from ..fluxoid import make_fluxoid_polygons

            points = make_fluxoid_polygons(self.device, holes=hole_name)[hole_name]
        hole = {hole.name: hole for hole in self.device.holes}[hole_name]
        if not Polygon(points=points).contains_points(hole.points).all():
            raise ValueError(
                f"Hole {hole_name} is not completely enclosed by the given polygon."
            )
        return self.polygon_fluxoid(
            points,
            interp_method=interp_method,
            units=units,
            with_units=with_units,
        )

    def boundary_phases(
        self, delta: bool = False
    ) -> Dict[str, Tuple[np.ndarray, np.ndarray]]:
        """Returns a dict of ``{polygon_name: (boundary_indices, boundary_phases)}``.

        ``(boundary_phases[-1] - boundary_phases[0]) / (2 * np.pi)`` gives the winding
        number for the polygon, i.e., the fluxoid in units of ``Phi_0``.

        .. seealso::

            :meth:`tdgl.Solution.hole_fluxoid`

        Args:
            delta: If True, ``boundary_phases[0]`` will be subtracted for each polygon.

        Returns:
            ``{polygon_name: (boundary_indices, boundary_phases)}``
        """
        device = self.device
        boundary_indices = device.boundary_sites()
        psi = self.tdgl_data.psi
        theta = np.angle(psi)
        phases = {}
        for name, indices in boundary_indices.items():
            phase = np.unwrap(theta[indices])
            if delta:
                phase -= phase[0]
            phases[name] = BoundaryPhases(indices, phase)
        return phases

    def current_through_path(
        self,
        path_coords: np.ndarray,
        dataset: Union[str, None] = None,
        method: str = "linear",
        units: Union[str, None] = None,
        with_units: bool = True,
    ) -> Union[float, pint.Quantity]:
        """Calculates the total current crossing a given path.

        Args:
            path_coords: An ``(n, 2)`` array of ``(x, y)`` coordinates defining
                the path.
            dataset: ``None``, ``"supercurrent"``, or ``"normal_current"``.
                ``None`` indicates the total current.
            method: Interpolation method: either "linear" or "cubic".
            units: The current units to return.
            with_units: Whether to return a :class:`pint.Quantity` with units attached.

        Returns:
            The total current crossing the path as either a float or a
            :class:`pint.Quantity`.
        """
        device = self.device
        if units is None:
            units = self.current_units
        J = self.interp_current_density(
            path_coords,
            dataset=dataset,
            method=method,
            with_units=True,
        )
        # The center of each edge in the path
        edge_positions = (path_coords[:-1] + path_coords[1:]) / 2
        # Evaluate the supercurrent at the edge centers
        J_edge = (J[:-1] + J[1:]) / 2
        edge_lengths, unit_normals = path_vectors(path_coords)
        edge_lengths = edge_lengths * device.ureg(device.length_units)
        J_dot_n = (J_edge * unit_normals).sum(axis=1)
        # Exclude points that are not inside the device.
        in_device = self.device.contains_points(edge_positions)
        total_current = np.trapz((J_dot_n * edge_lengths)[in_device]).to(units)
        if not with_units:
            total_current = total_current.magnitude
        return total_current

    def field_at_position(
        self,
        positions: np.ndarray,
        *,
        zs: Optional[Union[float, np.ndarray]] = None,
        vector: bool = False,
        units: Optional[str] = None,
        with_units: bool = True,
        return_sum: bool = True,
    ) -> Union[BiotSavartField, np.ndarray]:
        """Calculates the field due to currents in the device at any point(s) in space.

        .. seealso::

            :class:`tdgl.BiotSavartField`

        Args:
            positions: Shape (m, 2) array of (x, y) coordinates, or (m, 3) array
                of (x, y, z) coordinates at which to calculate the magnetic field.
                A single sequence like [x, y] or [x, y, z] is also allowed.
            zs: z coordinates at which to calculate the field. If positions has shape
                (m, 3), then this argument is not allowed. If zs is a scalar, then
                the fields are calculated in a plane parallel to the x-y plane.
                If zs is any array, then it must be same length as positions.
            vector: Whether to return the full vector magnetic field
                or just the z component.
            units: Units to which to convert the fields (can be either magnetic field H
                or magnetic flux density B = mu0 * H). If not given, then the fields
                are returned in units of ``self.field_units``.
            with_units: Whether to return the fields as ``pint.Quantity``
                with units attached.
            return_sum: If ``False``, this method will return a :class:`tdgl.BiotSavartField`
                instance, where the field from the supercurrent and normal current
                are identified separately.

        Returns:
            An np.ndarray if ``return_sum`` is ``True``, otherwise an instance of
            :class:`tdgl.BiotSavartField`. If ``with_units`` is ``True``, then the
            array(s) will be of type :class:`pint.Quantity`. The array(s) will have
            shape ``(m, )`` if vector is False, or shape ``(m, 3)`` if ``vector`` is True.
        """
        device = self.device
        ureg = device.ureg
        points = device.points
        units = units or self.field_units
        # In case something like a list [x, y] or [x, y, z] is given
        positions = np.atleast_2d(positions)
        # If positions includes z coordinates, peel those off here
        if positions.shape[1] == 3:
            if zs is not None:
                raise ValueError(
                    "If positions has shape (m, 3) then zs cannot be specified."
                )
            zs = positions[:, 2]
            positions = positions[:, :2]
        elif isinstance(zs, (int, float, np.generic)):
            # constant zs
            zs = zs * np.ones(len(positions))
        zs = zs.squeeze()
        if not isinstance(zs, np.ndarray):
            raise ValueError(f"Expected zs to be an ndarray, but got {type(zs)}.")
        weights = device.mesh.areas * device.coherence_length.magnitude**2
        # Compute the fields at the specified positions from the currents in each layer
        layer = self.device.layer
        if np.all((zs - layer.z0) == 0):
            if device.film.contains_points(positions).any():
                raise ValueError("Cannot interpolate fields within a film.")
        fields = []
        for name in ("supercurrent_density", "normal_current_density"):
            J = (
                getattr(self, name)
                .to(f"{self.current_units} / {device.length_units}")
                .magnitude
            )
            H = biot_savart_2d(
                positions[:, 0],
                positions[:, 1],
                zs,
                positions=points,
                areas=weights,
                current_densities=J,
                z0=layer.z0,
                length_units=device.length_units,
                current_units=self.current_units,
                vector=vector,
            )
            field = convert_field(
                H,
                units,
                old_units="tesla",
                ureg=ureg,
                with_units=with_units,
            )
            fields.append(field)
        fields = BiotSavartField(*fields)
        if return_sum:
            return sum(fields)
        return fields

    def vector_potential_at_position(
        self,
        positions: np.ndarray,
        *,
        zs: Union[float, np.ndarray, None] = None,
        units: Union[str, None] = None,
        with_units: bool = True,
        return_sum: bool = True,
    ) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        """Calculates the vector potential due to currents in the device at any
        point(s) in space, plus the applied vector potential.

        The vector potential :math:`\\mathbf{A}` at position :math:`\\mathbf{r}`
        due to sheet current density :math:`\\mathbf{K}(\\mathbf{r}')` flowing in a film
        with lateral geometry :math:`S` is:

        .. math::

            \\mathbf{A}(\\mathbf{r}) = \\frac{\\mu_0}{4\\pi}
            \\int_S\\frac{\\mathbf{K}(\\mathbf{r}')}{|\\mathbf{r}-\\mathbf{r}'|}\\mathrm{d}^2r'.

        Args:
            positions: Shape (m, 2) array of (x, y) coordinates, or (m, 3) array
                of (x, y, z) coordinates at which to calculate the vector potential.
                A single list like [x, y] or [x, y, z] is also allowed.
            zs: z coordinates at which to calculate the potential. If positions has shape
                (m, 3), then this argument is not allowed. If zs is a scalar, then
                the fields are calculated in a plane parallel to the x-y plane.
                If zs is any array, then it must be same length as positions.
            units: Units to which to convert the vector potential.
            with_units: Whether to return the vector potential as a ``pint.Quantity``
                with units attached.
            return_sum: Whether to return the total potential or a dict with keys
                ``("applied", "supercurrent", "normal_current")``.

        Returns:
            An np.ndarray if ``return_sum`` is ``True``, otherwise a dict of
            ``{source: potential_from_source}``. If ``with_units`` is ``True``, then
            the array(s) will be of type :class:`pint.Quantity`.
            ``potential_from_source`` will have shape ``(m, 3)``.
        """
        device = self.device
        ureg = device.ureg
        points = device.points
        areas = device.mesh.areas * device.coherence_length.magnitude**2
        units = units or f"{self.field_units} * {device.length_units}"
        # In case something like a list [x, y] or [x, y, z] is given
        positions = np.atleast_2d(positions)
        # If positions includes z coordinates, peel those off here
        if positions.shape[1] == 3:
            if zs is not None:
                raise ValueError(
                    "If positions has shape (m, 3) then zs cannot be specified."
                )
            zs = positions[:, 2]
            positions = positions[:, :2]
        elif isinstance(zs, (int, float, np.generic)):
            # constant zs
            zs = zs * np.ones(len(positions))
        if not isinstance(zs, np.ndarray):
            raise ValueError(f"Expected zs to be an ndarray, but got {type(zs)}.")
        if zs.ndim == 1:
            # We need zs to be shape (m, 1)
            zs = zs[:, np.newaxis]
        rho2 = distance.cdist(positions, points, metric="sqeuclidean")
        layer = device.layer
        vector_potentials = {}
        applied = self.applied_vector_potential(
            positions[:, 0],
            positions[:, 1],
            zs.squeeze(),
        )
        if applied.shape[1] == 2:
            applied = A = np.concatenate(
                [applied, np.zeros_like(applied[:, :1])], axis=1
            )
        applied = (applied * ureg(f"{self.field_units} * {device.length_units}")).to(
            units
        )
        if not with_units:
            applied = applied.magnitude
        vector_potentials["applied"] = applied
        dz = zs - layer.z0
        # rho has units of [length] and
        # shape = (postitions.shape[0], device.points.shape[0], 1)
        rho = np.sqrt(rho2 + dz**2)[:, :, np.newaxis]
        J_units = f"{self.current_units} / {device.length_units}"
        for name in ("supercurrent_density", "normal_current_density"):
            # J has units of [current / length], shape = (device.points.shape[0], 2)
            J = getattr(self, name).to(J_units).magnitude
            Axy = np.einsum("ijk, j -> ik", J / rho, areas)
            # z-component is zero because currents are parallel to the x-y plane.
            A = np.concatenate([Axy, np.zeros_like(Axy[:, :1])], axis=1)
            A = A * ureg(self.current_units)
            A = (ureg("mu_0") / (4 * np.pi) * A).to(units)
            if not with_units:
                A = A.magnitude
            vector_potentials[name] = A
        if return_sum:
            return sum(vector_potentials.values())
        return vector_potentials

    def _save_to_hdf5_file(
        self,
        h5file: Union[h5py.File, str],
        save_tdgl_data: bool = False,
        save_mesh: bool = True,
    ) -> None:
        def serialize_func(func, name, h5group):
            try:
                h5group.attrs[name] = func
            except TypeError:
                # Unsupported dtype - just pickle it.
                h5group[f"{name}.pickle"] = np.void(cloudpickle.dumps(func))

        if isinstance(h5file, str):
            mode = "x" if save_tdgl_data else "r+"
            save_context = h5py.File(h5file, mode, libver="latest")
        else:
            save_context = nullcontext(h5file)

        with save_context as f:
            if "mesh" in f:
                del f["mesh"]
            data_grp = f.require_group("data")
            if save_tdgl_data:
                self.tdgl_data.to_hdf5(data_grp)
                self.dynamics.to_hdf5(data_grp.require_group(str(self.tdgl_data.step)))
            if "solution" in f:
                del f["solution"]
            group = f.create_group("solution")
            options_grp = group.create_group("options")
            for k, v in dataclasses.asdict(self.options).items():
                if v is not None:
                    options_grp.attrs[k] = v
            group.attrs["time_created"] = self.time_created.isoformat()
            group.attrs["current_units"] = self.current_units
            group.attrs["field_units"] = self.field_units
            serialize_func(
                self.applied_vector_potential,
                "applied_vector_potential",
                group,
            )
            serialize_func(
                self.terminal_currents,
                "terminal_currents",
                group,
            )
            serialize_func(
                self.disorder_epsilon,
                "disorder_epsilon",
                group,
            )
            group.attrs["total_seconds"] = self.total_seconds
            self.device.to_hdf5(group.create_group("device"), save_mesh=save_mesh)

    def to_hdf5(self, h5path: Union[str, None] = None, save_mesh: bool = True) -> None:
        """Save the Solution to the existing output HDF5 file or to a new HDF5 file.

        Args:
            h5path: Path to an HDF5 file. If ``None`` is given, the
                :class:`tdgl.Solution` will be saved to the existing HDF5 output file
                located at ``self.path``.
            save_mesh: Whether to save the Device's mesh.
        """
        if self.saved_on_disk:
            if h5path is None:
                self._save_to_hdf5_file(self.path, save_mesh=save_mesh)
            else:
                shutil.copy(self.path, h5path)
                self._save_to_hdf5_file(h5path, save_mesh=save_mesh)
            return

        if h5path is None:
            raise ValueError(
                "The solution HDF5 file does not exist, "
                "and a new HDF5 file was not given."
            )
        self._save_to_hdf5_file(h5path, save_tdgl_data=True, save_mesh=save_mesh)

    @staticmethod
    def from_hdf5(path: os.PathLike, solve_step: int = -1) -> "Solution":
        """Loads a :class:`tdgl.Solution` from file.

        Args:
            path: Path to the HDF5 file containing a serialized :class:`tdgl.Solution`.
            solve_step: The solve step to load.

        Returns:
            The loaded Solution instance.
        """

        def deserialize_func(name, h5group):
            if name in h5group.attrs:
                return h5group.attrs[name]
            if f"{name}.pickle" in h5group:
                return pickle.loads(np.void(grp[f"{name}.pickle"]).tobytes())
            raise IOError(f"Unable to load {name}.")

        with h5py.File(path, "r", libver="latest") as f:
            grp = f["solution"]
            options_kwargs = dict()
            for k, v in grp["options"].attrs.items():
                options_kwargs[k] = v
            options = SolverOptions(**options_kwargs)
            time_created = datetime.fromisoformat(grp.attrs["time_created"])
            vector_potential = deserialize_func("applied_vector_potential", grp)
            terminal_currents = deserialize_func("terminal_currents", grp)
            disorder_epsilon = deserialize_func("disorder_epsilon", grp)
            total_seconds = grp.attrs["total_seconds"]
            device = Device.from_hdf5(grp["device"])

        solution = Solution(
            device=device,
            path=path,
            options=options,
            applied_vector_potential=vector_potential,
            terminal_currents=terminal_currents,
            disorder_epsilon=disorder_epsilon,
            total_seconds=total_seconds,
            _solve_step=solve_step,
        )
        solution._time_created = time_created
        return solution

    def delete_hdf5(self) -> None:
        """Delete the HDF5 file accompanying the :class:`tdgl.Solution`."""
        if self.saved_on_disk:
            os.remove(self.path)

    def equals(
        self,
        other: Any,
        require_same_timestamp: bool = False,
    ) -> bool:
        """Checks whether two solutions are equal.

        Args:
            other: The :class:`tdgl.Solution` to compare for equality.
            require_same_timestamp: If True, two solutions are only considered
                equal if they have the exact same time_created.

        Returns:
            A boolean indicating whether the two solutions are equal
        """
        # First check things that are "easy" to check
        if other is self:
            return True
        if not isinstance(other, Solution):
            return False

        def compare_callables(first, second):
            if isinstance(first, Parameter):
                return first == second
            if callable(first):
                if not callable(second):
                    return False
                get_code = operator.attrgetter("co_code", "co_consts")
                if get_code(first.__code__) != get_code(second.__code__):
                    return False
            elif first != second:
                return False
            return True

        if not (
            (self.device == other.device)
            and (self.options == other.options)
            and (self.solve_step == other.solve_step)
            and compare_callables(
                self.applied_vector_potential, other.applied_vector_potential
            )
            and compare_callables(self.terminal_currents, other.terminal_currents)
            and compare_callables(self.disorder_epsilon, other.disorder_epsilon)
            and (self.tdgl_data == other.tdgl_data)
            and (self.dynamics == other.dynamics)
        ):
            return False
        if require_same_timestamp and (self.time_created != other.time_created):
            return False
        return True

    def __eq__(self, other) -> bool:
        return self.equals(other, require_same_timestamp=True)

    def plot_currents(self, **kwargs) -> Tuple[plt.Figure, np.ndarray]:
        """An alias for :func:`tdgl.plot_currents`."""
        from .plot_solution import plot_currents

        return plot_currents(self, **kwargs)

    def plot_order_parameter(self, **kwargs) -> Tuple[plt.Figure, np.ndarray]:
        """An alias for :func:`tdgl.plot_order_parameter`."""
        from .plot_solution import plot_order_parameter

        return plot_order_parameter(self, **kwargs)

    def plot_field_at_positions(
        self, positions: np.ndarray, **kwargs
    ) -> Tuple[plt.Figure, np.ndarray]:
        """An alias for :func:`tdgl.plot_field_at_positions`."""
        from .plot_solution import plot_field_at_positions

        return plot_field_at_positions(self, positions, **kwargs)

    def plot_vorticity(self, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """An alias for :func:`tdgl.plot_vorticity`."""
        from .plot_solution import plot_vorticity

        return plot_vorticity(self, **kwargs)

    def plot_scalar_potential(self, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """An alias for :func:`tdgl.plot_scalar_potential`."""
        from .plot_solution import plot_scalar_potential

        return plot_scalar_potential(self, **kwargs)



import itertools
import logging
import os
import tempfile
import time
import traceback
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Sequence, Tuple, Union

import h5py
import numpy as np
from tqdm import TqdmWarning, tqdm

from ..finite_volume.mesh import Mesh
from .options import SolverOptions


class DataHandler:
    """A context manager that is responsible for reading from and writing to disk.

    Args:
        output_file: File to use as output for simulation data.
        logger: Logger used to inform about errors.
    """

    def __init__(
        self,
        output_file: Union[str, None],
        logger: Union[logging.Logger, None] = None,
    ):
        self.output_file = None
        self.tempdir = None
        self.mesh_group = None
        self.time_step_group = None
        self.save_number = 0
        self.logger = logger if logger is not None else logging.getLogger()
        self._base_output_file = output_file
        self.output_file: Union[h5py.File, None] = None
        self.output_path: Union[str, None] = None
        self.time_step_group: Union[h5py.Group, None] = None
        self.mesh_group: Union[h5py.Group, None] = None

    def _create_output_file(self, output: str) -> Tuple[h5py.File, str]:
        """Create an output file.

        Args:
            output: The output file path.
            logger: Logger output logs.

        Returns:
            A :class:`h5py.File` and the path to the file.
        """
        if output is None:
            self.tempdir = tempfile.TemporaryDirectory()
            directory = self.tempdir.name
            name = "output"
            suffix = "h5"
        else:
            # Make sure the directory exists
            Path(output).parent.mkdir(parents=True, exist_ok=True)
            # Split the output into file name and suffix
            name_parts = output.split(".")
            name = ".".join(name_parts[:-1])
            suffix = name_parts[-1]
            directory = os.getcwd()
        # Number to be added to the end of file name
        # If this is None do not add the number
        serial_number = None
        while True:
            # Create a new file name
            name_suffix = f"-{serial_number}" if serial_number is not None else ""
            file_name = f"{name}{name_suffix}.{suffix}"
            file_path = os.path.join(directory, file_name)
            try:
                file = h5py.File(file_path, "x", libver="latest")
            except (OSError, FileExistsError):
                if serial_number is None:
                    serial_number = 1
                else:
                    serial_number += 1
                continue
            else:
                if serial_number is not None:
                    self.logger.warning(
                        f"Output file already exists. Renaming to {file_name}."
                    )
            return file, file_path

    def __enter__(self) -> "DataHandler":
        self.output_file, self.output_path = self._create_output_file(
            self._base_output_file
        )
        self.time_step_group = self.output_file.create_group("data", track_order=True)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        if exc_value is not None:
            self.logger.warning(
                "Ignoring the following exception in DataHandler.__exit__():"
            )
            self.logger.warning(
                "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
            )
        self.close()

    def close(self):
        """Clean up by closing the output file."""
        self.output_file.close()
        if self.tempdir is not None:
            self.tempdir.cleanup()

    def save_mesh(self, mesh: Mesh) -> None:
        """Save the ``Mesh`` to ``self.output_file["mesh"]``.

        Args:
            mesh: The ``Mesh`` to save.
        """
        self.mesh_group = self.output_file.create_group("mesh")
        mesh.to_hdf5(self.mesh_group)

    def save_fixed_values(self, fixed_data: Dict[str, np.ndarray]) -> None:
        """Save the fixed values, i.e., those that aren't updated at each solve step."""
        for key, value in fixed_data.items():
            self.output_file[key] = value

    def save_time_step(
        self,
        state: Dict[str, float],
        data: Dict[str, np.ndarray],
        running_state: Union[Dict[str, np.ndarray], None],
    ) -> None:
        """Save the state and data that are updated at each solve step."""
        group = self.time_step_group.create_group(f"{self.save_number}")
        group.attrs["timestamp"] = datetime.now().isoformat()
        self.save_number += 1
        for key, value in state.items():
            group.attrs[key] = value
        for key, value in data.items():
            group[key] = value
        if running_state is not None:
            running_grp = group.create_group("running_state")
            for key, value in running_state.items():
                running_grp[key] = np.squeeze(value)


class RunningState:
    """Storage class for saving scalar data that is saved at each time step.

    Args:
        names_and_sizes: A dict of the parameters to be saved and their sizes, i.e.,
            the number of values measured at each time step for each parameter.
        buffer_size: The size of the buffer, i.e., the number of values to record
            before writing to disk.
    """

    def __init__(self, names_and_sizes: Dict[str, int], buffer_size: int):
        self.step = 0
        self.buffer_size = buffer_size
        self.names_and_sizes = names_and_sizes
        self.values = {
            name: np.zeros((size, buffer_size))
            for name, size in self.names_and_sizes.items()
        }

    def clear(self) -> None:
        """Clear the buffer."""
        self.step = 0
        for name, size in self.names_and_sizes.items():
            self.values[name] = np.zeros((size, self.buffer_size))

    def append(self, name: str, value: Sequence[float]) -> None:
        """Append data to the buffer.

        Args:
            name: Data to append.
            value: Value of the data.
        """
        self.values[name][:, self.step] = value


class Runner:
    """The runner is responsible for the simulation loop. It handles the state,
    runs the update function and saves the data.

    Args:
        function: The update function that takes the state from the current state
            to the next.
        options: SolverOptions instance.
        initial_values: Initial values passed as parameters to the update function.
        names: Names of the parameters.
        data_handler: The data handler used to save to disk.
        fixed_values: Values that do not change over time, but should be added
            to saved data.
        fixed_names: Fixed data variable names.
        running_names_and_sizes: Names and shapes of running state variables.
        logger: A logger to print information about simulation.
        state: The current state variables.
    """

    def __init__(
        self,
        function: Callable,
        options: SolverOptions,
        initial_values: List[Any],
        names: Sequence[str],
        data_handler: DataHandler,
        fixed_values: Union[List[Any], None] = None,
        fixed_names: Union[Sequence, None] = None,
        running_names_and_sizes: Union[Dict[str, int], None] = None,
        logger: Union[logging.Logger, None] = None,
        state: Union[Dict[str, Any], None] = None,
    ):
        self.time = 0
        self.options = options
        self.dt = self.options.dt_init
        self.function = function
        self.values = initial_values
        self.names = names
        self.fixed_values = fixed_values if fixed_values is not None else []
        self.fixed_names = fixed_names if fixed_names is not None else []
        self.running_names_and_sizes = (
            running_names_and_sizes if running_names_and_sizes is not None else {}
        )
        self.running_state = RunningState(
            self.running_names_and_sizes, self.options.save_every
        )
        self.state = state if state is not None else {}
        self.logger = logger if logger is not None else logging.getLogger()
        self.data_handler = data_handler

    def run(self) -> None:
        """Run the simulation loop."""
        self.time = 0
        self.state["step"] = 0
        self.state["time"] = self.time
        self.state["dt"] = self.dt
        self.data_handler.save_fixed_values(
            dict(zip(self.fixed_names, self.fixed_values))
        )
        # Thermalize if enabled
        if self.options.skip_time:
            self._run_stage_(
                "Thermalizing",
                start_time=self.time,
                end_time=self.options.skip_time,
                save=False,
            )
            self.running_state.clear()
        self.time = 0
        self.state["step"] = 0
        self.state["time"] = self.time
        self.state["dt"] = self.options.dt_init
        # Run the simulation
        self._run_stage_(
            "Simulating",
            start_time=self.time,
            end_time=self.options.solve_time,
            save=True,
        )

    def _run_stage_(
        self,
        name: str,
        start_time: float,
        end_time: float,
        save: bool = True,
    ) -> None:
        """Run a stage of the simulation.

        Args:
            name: Name of the solver stage.
            start_time: Start time.
            end_time: End time.
            save: If the data should be saved.
        """
        prog_disabled = (
            self.options.progress_interval is not None
            and self.options.progress_interval > 0
        )
        now = None
        initial = start_time
        total = end_time
        r_bar = "| {n:.0f}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt} {postfix}]"
        unit = "tau"
        bar_format = "{l_bar}{bar}" + r_bar
        it = itertools.count()

        def save_step(step):
            data = dict(zip(self.names, self.values))
            if step == 0:
                running_state = None
            else:
                running_state = self.running_state.values
            self.data_handler.save_time_step(self.state, data, running_state)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=TqdmWarning)
            with tqdm(
                initial=initial,
                total=total,
                desc=name,
                disable=prog_disabled,
                unit=unit,
                bar_format=bar_format,
                dynamic_ncols=True,
            ) as pbar:
                for i in it:
                    self.state["step"] = i
                    self.state["time"] = self.time
                    self.state["dt"] = self.dt
                    # Print progress if TQDM is disabled.
                    if prog_disabled and (i % self.options.progress_interval) == 0:
                        then, now = now, time.perf_counter()
                        if then is None:
                            speed = 0
                        else:
                            speed = self.options.progress_interval / (now - then)
                        self.logger.info(
                            f"{name}: Time {self.time}/{end_time}, "
                            f"dt={self.dt:.2e}, {speed:.2f} it/s"
                        )
                    if i % self.options.save_every == 0:
                        if save:
                            save_step(i)
                        self.running_state.clear()
                    # Run time step.
                    function_result = self.function(
                        self.state,
                        self.running_state,
                        *self.values,
                        self.dt,
                    )
                    *self.values, new_dt = function_result
                    # tqdm will spit out a warning if you try to update past "total"
                    if self.time + self.dt < end_time:
                        pbar.update(self.dt)
                    else:
                        pbar.update(end_time - self.time)
                    if self.time >= end_time:
                        break
                    self.dt = new_dt
                    self.running_state.step += 1
                    self.time += self.dt
            if save:
                save_step(i)



import itertools
import logging
from datetime import datetime
from typing import Callable, Dict, Sequence, Tuple, Union

import numpy as np
import scipy.sparse as sp
from scipy import spatial

try:
    import jax
    import jax.numpy as jnp
except (ModuleNotFoundError, ImportError):
    jax = None

from ..device.device import Device, TerminalInfo
from ..em import ureg
from ..finite_volume.operators import MeshOperators
from ..finite_volume.util import get_supercurrent
from ..solution.solution import Solution
from ..sources.constant import ConstantField
from .options import SolverOptions
from .runner import DataHandler, Runner

logger = logging.getLogger(__name__)


if jax is None:
    einsum = np.einsum
else:
    einsum = jnp.einsum


def validate_terminal_currents(
    terminal_currents: Union[Callable, Dict[str, float]],
    terminal_info: Sequence[TerminalInfo],
    solver_options: SolverOptions,
    num_evals: int = 100,
) -> None:
    """Ensure that the terminal currents satisfy current conservation."""

    def check_total_current(currents: Dict[str, float]):
        names = set([t.name for t in terminal_info])
        if unknown := set(currents).difference(names):
            raise ValueError(
                f"Unknown terminal(s) in terminal currents: {list(unknown)}."
            )
        total_current = sum(currents.values())
        if total_current:
            raise ValueError(
                f"The sum of all terminal currents must be 0 (got {total_current:.2e})."
            )

    if callable(terminal_currents):
        times = np.random.default_rng().random(num_evals) * solver_options.solve_time
        for t in times:
            check_total_current(terminal_currents(t))
    else:
        check_total_current(terminal_currents)


def solve_for_psi_squared(
    *,
    psi: np.ndarray,
    abs_sq_psi: np.ndarray,
    mu: np.ndarray,
    epsilon: np.ndarray,
    gamma: float,
    u: float,
    dt: float,
    psi_laplacian: sp.spmatrix,
) -> Union[Tuple[np.ndarray, np.ndarray], None]:
    """Solve for :math:`\\psi_i^{n+1}` and :math:`|\\psi_i^{n+1}|^2` given
    :math:`\\psi_i^n` and :math:`\\mu_i^n`.

    Args:
        psi: The current value of the order parameter, :math:`\\psi_^n`
        abs_sq_psi: The current value of the superfluid density, :math:`|\\psi^n|^2`
        mu: The current value of the electric potential, :math:`\\mu^n`
        epsilon: The disorder parameter, :math:`\\epsilon`
        gamma: The inelastic scattering parameter, :math:`\\gamma`.
        u: The ratio of relaxation times for the order parameter, :math:`u`
        dt: The time step
        psi_laplacian: The covariant Laplacian for the order parameter

    Returns:
        ``None`` if the calculation failed to converge, otherwise the new order
        parameter :math:`\\psi^{n+1}` and superfluid density :math:`|\\psi^{n+1}|^2`.
    """
    U = np.exp(-1j * mu * dt)
    z = U * gamma**2 / 2 * psi
    with np.errstate(all="raise"):
        try:
            w = z * abs_sq_psi + U * (
                psi
                + (dt / u)
                * np.sqrt(1 + gamma**2 * abs_sq_psi)
                * ((epsilon - abs_sq_psi) * psi + psi_laplacian @ psi)
            )
            c = w.real * z.real + w.imag * z.imag
            two_c_1 = 2 * c + 1
            w2 = np.abs(w) ** 2
            discriminant = two_c_1**2 - 4 * np.abs(z) ** 2 * w2
        except Exception:
            logger.debug("Unable to solve for |psi|^2.", exc_info=True)
            return None
    if np.any(discriminant < 0):
        return None
    new_sq_psi = (2 * w2) / (two_c_1 + np.sqrt(discriminant))
    psi = w - z * new_sq_psi
    return psi, new_sq_psi


def adaptive_euler_step(
    step: int,
    psi: np.ndarray,
    abs_sq_psi: np.ndarray,
    mu: np.ndarray,
    epsilon: np.ndarray,
    gamma: float,
    u: float,
    dt: float,
    psi_laplacian: MeshOperators,
    options: SolverOptions,
) -> Tuple[np.ndarray, np.ndarray, float]:
    """Update the order parameter and time step in an adaptive Euler step.

    Args:
        step: The solve step index, :math:`n`
        psi: The current value of the order parameter, :math:`\\psi_^n`
        abs_sq_psi: The current value of the superfluid density, :math:`|\\psi^n|^2`
        mu: The current value of the electric potential, :math:`\\mu^n`
        epsilon: The disorder parameter, :math:`\\epsilon`
        gamma: The inelastic scattering parameter, :math:`\\gamma`.
        u: The ratio of relaxation times for the order parameter, :math:`u`
        dt: The tentative time step, which will be updated
        psi_laplacian: The covariant Laplacian for the order parameter
        options: The solver options for the simulation

    Returns:
        :math:`\\psi^{n+1}`, :math:`|\\psi^{n+1}|^2`, and :math:`\\Delta t^{n}`.
    """
    kwargs = dict(
        psi=psi,
        abs_sq_psi=abs_sq_psi,
        mu=mu,
        epsilon=epsilon,
        gamma=gamma,
        u=u,
        dt=dt,
        psi_laplacian=psi_laplacian,
    )
    result = solve_for_psi_squared(**kwargs)
    for retries in itertools.count():
        if result is not None:
            break  # First evaluation of |psi|^2 was successful.
        if (not options.adaptive) or retries > options.max_solve_retries:
            raise RuntimeError(
                f"Solver failed to converge in {options.max_solve_retries}"
                f" retries at step {step} with dt = {dt:.2e}."
                f" Try using a smaller dt_init."
            )
        kwargs["dt"] = dt = dt * options.adaptive_time_step_multiplier
        result = solve_for_psi_squared(**kwargs)
    psi, new_sq_psi = result
    return psi, new_sq_psi, dt


def solve(
    device: Device,
    options: SolverOptions,
    applied_vector_potential: Union[Callable, float] = 0,
    terminal_currents: Union[Callable, Dict[str, float], None] = None,
    disorder_epsilon: Union[float, Callable] = 1,
    seed_solution: Union[Solution, None] = None,
):
    """Solve a TDGL model.

    Args:
        device: The :class:`tdgl.Device` to solve.
        options: An instance :class:`tdgl.SolverOptions` specifying the solver
            parameters.
        applied_vector_potential: A function or :class:`tdgl.Parameter` that computes
            the applied vector potential as a function of position ``(x, y, z)``. If a float
            ``B`` is given, the applied vector potential will be that of a uniform magnetic
            field with strength ``B`` ``field_units``.
        terminal_currents: A dict of ``{terminal_name: current}`` or a callable with signature
            ``func(time: float) -> {terminal_name: current}``, where ``current`` is a float
            in units of ``current_units`` and ``time`` is the dimensionless time.
        disorder_epsilon: A float in range [-1, 1], or a callable with signature
            ``disorder_epsilon(r: Tuple[float, float]) -> epsilon``, where ``epsilon``
            is a float in range [-1, 1]. Setting
            :math:`\\epsilon(\\mathbf{r})=T_c(\\mathbf{r})/T_c - 1 < 1` suppresses the
            critical temperature at position :math:`\\mathbf{r}`, which can be used
            to model inhomogeneity.
        seed_solution: A :class:`tdgl.Solution` instance to use as the initial state
            for the simulation.

    Returns:
        A :class:`tdgl.Solution` instance.
    """

    start_time = datetime.now()
    options.validate()
    current_units = options.current_units
    field_units = options.field_units
    output_file = options.output_file

    mesh = device.mesh
    sites = device.points
    edges = mesh.edge_mesh.edges
    length_units = ureg(device.length_units)
    xi = device.coherence_length.magnitude
    probe_points = device.probe_point_indices
    u = device.layer.u
    gamma = device.layer.gamma
    K0 = device.K0
    # The vector potential is evaluated on the mesh edges,
    # where the edge coordinates are in dimensionful units.
    x = mesh.edge_mesh.x * xi
    y = mesh.edge_mesh.y * xi
    Bc2 = device.Bc2
    z = device.layer.z0 * xi * np.ones_like(x)
    J_scale = 4 * ((ureg(current_units) / length_units) / K0).to_base_units()
    assert "dimensionless" in str(J_scale.units), str(J_scale.units)
    J_scale = J_scale.magnitude
    if not callable(applied_vector_potential):
        applied_vector_potential = ConstantField(
            applied_vector_potential,
            field_units=field_units,
            length_units=device.length_units,
        )
    # Evaluate the vector potential
    vector_potential = applied_vector_potential(x, y, z)
    vector_potential = np.asarray(vector_potential)[:, :2]
    shape = vector_potential.shape
    if shape != x.shape + (2,):
        raise ValueError(f"Unexpected shape for vector_potential: {shape}.")
    vector_potential = (
        (vector_potential * ureg(field_units) * length_units)
        / (Bc2 * xi * length_units)
    ).to_base_units()
    assert "dimensionless" in str(vector_potential.units)
    vector_potential = vector_potential.magnitude

    dt_max = options.dt_max if options.adaptive else options.dt_init

    # Find the current terminal sites.
    terminal_info = device.terminal_info()
    if terminal_info:
        normal_boundary_index = np.concatenate(
            [t.site_indices for t in terminal_info], dtype=np.int64
        )
    else:
        normal_boundary_index = np.array([], dtype=np.int64)
    # Define the source-drain current.
    if terminal_currents and device.probe_points is None:
        logger.warning(
            "The terminal currents are non-null, but the device has no probe points."
        )
    terminal_names = [term.name for term in terminal_info]
    if terminal_currents is None:
        terminal_currents = {name: 0 for name in terminal_names}
    if callable(terminal_currents):
        current_func = terminal_currents
    else:
        terminal_currents = {
            name: terminal_currents.get(name, 0) for name in terminal_names
        }

        def current_func(t):
            return terminal_currents

    validate_terminal_currents(current_func, terminal_info, options)

    # Construct finite-volume operators
    operators = MeshOperators(mesh, fixed_sites=normal_boundary_index)
    operators.build_operators()
    operators.set_link_exponents(vector_potential)
    divergence = operators.divergence
    mu_boundary_laplacian = operators.mu_boundary_laplacian
    mu_laplacian_lu = operators.mu_laplacian_lu
    mu_gradient = operators.mu_gradient
    # Initialize the order parameter and electric potential
    psi_init = np.ones(len(mesh.sites), dtype=np.complex128)
    psi_init[normal_boundary_index] = 0
    mu_init = np.zeros(len(mesh.sites))
    mu_boundary = np.zeros_like(mesh.edge_mesh.boundary_edge_indices, dtype=float)
    # Create the epsilon parameter, which is the maximum value of |psi| at each position.
    if callable(disorder_epsilon):
        epsilon = np.apply_along_axis(disorder_epsilon, 1, sites).astype(float)
    else:
        epsilon = disorder_epsilon * np.ones(len(sites))
    if np.any(epsilon < -1) or np.any(epsilon > 1):
        raise ValueError("The disorder parameter epsilon must be in range [-1, 1].")

    if options.include_screening:
        # Pre-compute the kernel for the screening integral.
        directions = mesh.edge_mesh.directions
        directions = directions / np.linalg.norm(directions, axis=1)[:, np.newaxis]
        inv_rho = 1 / spatial.distance.cdist(mesh.edge_mesh.centers, mesh.sites)
        # (edges, sites, spatial dimensions)
        inv_rho = inv_rho[:, :, np.newaxis] * mesh.areas[np.newaxis, :, np.newaxis]
        A_scale = (ureg("mu_0") / (4 * np.pi) * K0 / Bc2).to_base_units().magnitude
        inv_rho *= A_scale

        if jax is not None:
            # Even without a GPU, jax.numpy.einsum seems to be much faster
            # than numpy.einsum. This may just be because jax uses float32 by default.
            inv_rho = jax.device_put(inv_rho)

    # Running list of the max abs change in |psi|^2 between subsequent solve steps.
    # This list is used to calculate the adaptive time step.
    d_psi_sq_vals = []
    tentative_dt = options.dt_init
    # Parameters for the self-consistent screening calculation.
    step_size = options.screening_step_size
    drag = options.screening_step_drag

    # This is the function called at each step of the solver.
    def update(
        state,
        running_state,
        psi,
        mu,
        supercurrent,
        normal_current,
        A_induced,
        dt,
    ):
        nonlocal tentative_dt
        step = state["step"]
        time = state["time"]
        old_sq_psi = np.abs(psi) ** 2
        # Compute the current density for this step
        # and update the current boundary conditions.
        currents = current_func(time)
        for term in terminal_info:
            current_density = (-1 / term.length) * sum(
                currents.get(name, 0) for name in terminal_names if name != term.name
            )
            mu_boundary[term.boundary_edge_indices] = J_scale * current_density

        screening_error = np.inf
        A_induced_vals = []
        v = [0]  # Velocity for Polyak's method
        # This loop runs only once if options.include_screening is False
        for screening_iterations in itertools.count():
            if screening_error < options.screening_tolerance:
                break  # Screening calculation converged.
            if screening_iterations > options.max_iterations_per_step:
                raise RuntimeError(
                    f"Screening calculation failed to converge at step {step} after"
                    f" {options.max_iterations_per_step} iterations. Relative error in"
                    f" induced vector potential: {screening_error:.2e}"
                    f" (tolerance: {options.screening_tolerance:.2e})."
                )
            if options.include_screening:
                # If screening is included, update the link variables in the covariant
                # Laplacian and gradient for psi based on the induced vector potential
                # from the previous iteration.
                operators.set_link_exponents(vector_potential + A_induced)
            # Adjust the time step and calculate the new the order parameter
            if screening_iterations == 0:
                # Find a new time step only for the first screening iteration.
                dt = tentative_dt
            psi, abs_sq_psi, dt = adaptive_euler_step(
                step,
                psi,
                old_sq_psi,
                mu,
                epsilon,
                gamma,
                u,
                dt,
                operators.psi_laplacian,
                options,
            )
            # Compute the supercurrent, scalar potential, and normal current
            supercurrent = get_supercurrent(psi, operators.psi_gradient, edges)
            lhs = (divergence @ supercurrent) - (mu_boundary_laplacian @ mu_boundary)
            mu = mu_laplacian_lu.solve(lhs)
            normal_current = -(mu_gradient @ mu)

            if not options.include_screening:
                break

            # Evaluate the induced vector potential
            J_site = mesh.get_quantity_on_site(supercurrent + normal_current)
            # i: edges, j: sites, k: spatial dimensions
            new_A_induced = np.asarray(einsum("jk, ijk -> ik", J_site, inv_rho))
            # Update induced vector potential using Polyak's method
            dA = new_A_induced - A_induced
            v.append((1 - drag) * v[-1] + step_size * dA)
            A_induced = A_induced + v[-1]
            A_induced_vals.append(A_induced)
            if len(A_induced_vals) > 1:
                screening_error = np.max(
                    np.linalg.norm(dA, axis=1) / np.linalg.norm(A_induced, axis=1)
                )
                del v[:-2]
                del A_induced_vals[:-2]

        running_state.append("dt", dt)
        if probe_points is not None:
            # Update the voltage and phase difference
            running_state.append("mu", mu[probe_points])
            running_state.append("theta", np.angle(psi[probe_points]))
        if options.include_screening:
            running_state.append("screening_iterations", screening_iterations)

        if options.adaptive:
            # Compute the max abs change in |psi|^2, averaged over the adaptive window,
            # and use it to select a new time step.
            d_psi_sq_vals.append(np.abs(abs_sq_psi - old_sq_psi).max())
            window = options.adaptive_window
            if step > window:
                new_dt = options.dt_init / max(1e-10, np.mean(d_psi_sq_vals[-window:]))
                tentative_dt = np.clip(0.5 * (new_dt + dt), 0, dt_max)

        return (
            psi,
            mu,
            supercurrent,
            normal_current,
            A_induced,
            dt,
        )

    # Set the initial conditions.
    if seed_solution is None:
        num_edges = len(edges)
        parameters = {
            "psi": psi_init,
            "mu": mu_init,
            "supercurrent": np.zeros(num_edges),
            "normal_current": np.zeros(num_edges),
            "induced_vector_potential": np.zeros((num_edges, 2)),
        }
    else:
        if seed_solution.device != device:
            raise ValueError(
                "The seed_solution.device must be equal to the device being simulated."
            )
        seed_data = seed_solution.tdgl_data
        parameters = {
            "psi": seed_data.psi,
            "mu": seed_data.mu,
            "supercurrent": seed_data.supercurrent,
            "normal_current": seed_data.normal_current,
            "induced_vector_potential": seed_data.induced_vector_potential,
        }
    running_names_and_sizes = {"dt": 1}
    if probe_points is not None:
        running_names_and_sizes["mu"] = len(probe_points)
        running_names_and_sizes["theta"] = len(probe_points)
    if options.include_screening:
        running_names_and_sizes["screening_iterations"] = 1

    with DataHandler(output_file=output_file, logger=logger) as data_handler:
        data_handler.save_mesh(mesh)
        Runner(
            function=update,
            options=options,
            data_handler=data_handler,
            initial_values=list(parameters.values()),
            names=list(parameters),
            fixed_values=(vector_potential, epsilon),
            fixed_names=("applied_vector_potential", "epsilon"),
            running_names_and_sizes=running_names_and_sizes,
            logger=logger,
        ).run()
        end_time = datetime.now()
        logger.info(f"Simulation ended at {end_time}")
        logger.info(f"Simulation took {end_time - start_time}")

        solution = Solution(
            device=device,
            path=data_handler.output_path,
            options=options,
            applied_vector_potential=applied_vector_potential,
            terminal_currents=terminal_currents,
            disorder_epsilon=disorder_epsilon,
            total_seconds=(end_time - start_time).total_seconds(),
        )
        solution.to_hdf5()
    return solution



import numpy as np
import pytest

import tdgl
from tdgl.solver.options import SolverOptionsError


@pytest.mark.parametrize("current", [5.0, 10.0, lambda t: 10])
@pytest.mark.parametrize("field", [0, 1])
def test_source_drain_current(transport_device, current, field):

    device = transport_device
    total_time = 100

    options = tdgl.SolverOptions(
        solve_time=total_time,
        field_units="uT",
        current_units="uA",
        save_every=100,
    )
    if callable(current):

        def terminal_currents(t):
            return dict(source=current(0), drain=-current(0))

    else:
        terminal_currents = dict(source=current, drain=-current)
    solution = tdgl.solve(
        device,
        options,
        applied_vector_potential=field,
        terminal_currents=terminal_currents,
    )

    if callable(current):
        current = current(0)

    ys = np.linspace(-5, 5, 501)
    measured_currents = []
    for x0 in [-8, -2, 0, 2, 8]:
        coords = np.array([x0 * np.ones_like(ys), ys]).T
        measured_currents.append(
            solution.current_through_path(coords, with_units=False)
        )
    measured_currents = np.array(measured_currents)
    assert np.allclose(measured_currents, current, rtol=0.1)


def test_screening(box_device):
    device = box_device
    total_time = 50

    with pytest.raises(SolverOptionsError):
        tdgl.SolverOptions(solve_time=total_time, screening_step_size=-1).validate()
    with pytest.raises(SolverOptionsError):
        tdgl.SolverOptions(solve_time=total_time, screening_tolerance=-1).validate()
    with pytest.raises(SolverOptionsError):
        tdgl.SolverOptions(solve_time=total_time, screening_step_drag=2).validate()
    with pytest.raises(SolverOptionsError):
        tdgl.SolverOptions(
            solve_time=total_time, adaptive_time_step_multiplier=2
        ).validate()

    options = tdgl.SolverOptions(
        solve_time=total_time,
        field_units="uT",
        current_units="uA",
    )
    field = tdgl.sources.ConstantField(50)

    options.include_screening = False
    solution = tdgl.solve(
        device,
        options,
        applied_vector_potential=field,
    )

    circle = tdgl.geometry.circle(2, points=401)
    centers = [(0, 0), (-1.5, -2.5), (2.5, 2), (0, 1)]

    for r0 in centers:
        fluxoid = solution.polygon_fluxoid(circle + np.atleast_2d(r0), with_units=False)
        assert abs(sum(fluxoid) / fluxoid.flux_part) >= 2e-2

    options.include_screening = True
    solution = tdgl.solve(
        device,
        options,
        applied_vector_potential=field,
    )

    for r0 in centers:
        fluxoid = solution.polygon_fluxoid(circle + np.atleast_2d(r0), with_units=False)
        assert abs(sum(fluxoid) / fluxoid.flux_part) < 2e-2



import copy
import os
import pickle
import tempfile

import matplotlib.pyplot as plt
import numpy as np
import pint
import pytest
import shapely

import tdgl


def test_set_polygon_points():
    box = shapely.geometry.box(0, 0, 1, 1).exterior.coords
    hole = shapely.geometry.box(0.25, 0.25, 0.75, 0.75, ccw=False)
    polygon = shapely.geometry.polygon.Polygon(box, holes=[hole.exterior.coords])

    with pytest.raises(ValueError):
        _ = tdgl.Polygon("bad", points=polygon)

    invalid = shapely.geometry.polygon.LinearRing(
        [(0, 0), (0, 2), (1, 1), (2, 2), (2, 0), (1, 1), (0, 0)]
    )

    with pytest.raises(ValueError):
        _ = tdgl.Polygon(points=invalid)

    x, y = tdgl.geometry.circle(1).T
    z = np.ones_like(x)
    points = np.stack([x, y, z], axis=1)
    with pytest.raises(ValueError):
        _ = tdgl.Polygon(points=points)


def test_polygon_on_boundary(radius=1):
    points = tdgl.geometry.circle(radius, points=501)
    polygon = tdgl.Polygon(points=points)
    Delta_x, Delta_y = polygon.extents
    assert np.isclose(Delta_x, 2 * radius)
    assert np.isclose(Delta_y, 2 * radius)

    smaller = tdgl.geometry.circle(radius - 0.01)
    bigger = tdgl.geometry.circle(radius + 0.01)
    assert polygon.on_boundary(smaller, radius=0.1).all()
    assert polygon.on_boundary(bigger, radius=0.1).all()
    assert not polygon.on_boundary(smaller, radius=0.001).any()
    assert not polygon.on_boundary(bigger, radius=0.001).any()
    assert polygon.on_boundary(smaller, index=True).dtype is np.dtype(int)


def test_polygon_join():

    square1 = tdgl.Polygon(points=tdgl.geometry.box(1))
    square2 = tdgl.Polygon(points=tdgl.geometry.box(1)).translate(dx=0.5, dy=0.5)
    square3 = tdgl.geometry.box(1, center=(-0.25, 0.25))
    name = "name"
    for items in (
        [square1, square2, square3],
        [square1.points, square2.points, square3],
        [square1.polygon, square2.polygon, square3],
    ):
        _ = tdgl.Polygon.from_union(items, name=name)
        _ = tdgl.Polygon.from_difference(items, name=name)
        _ = tdgl.Polygon.from_intersection(items, name=name)

    assert (
        square1.union(square2, square3).polygon
        == tdgl.Polygon.from_union(items, name=name).polygon
    )

    assert (
        square1.union(square2, square3).polygon == (square1 + square2 + square3).polygon
    )

    assert (
        square1.intersection(square2, square3).polygon
        == tdgl.Polygon.from_intersection(items, name=name).polygon
    )

    assert (
        square1.intersection(square2, square3).polygon
        == (square1 * square2 * square3).polygon
    )

    assert (
        square1.difference(square2, square3).polygon
        == tdgl.Polygon.from_difference(items, name=name).polygon
    )

    assert (
        square1.difference(square2, square3).polygon
        == (square1 - square2 - square3).polygon
    )

    square1.layer = square2.layer = None
    with pytest.raises(ValueError):
        _ = tdgl.Polygon.from_difference([square1, square1], name=name)

    with pytest.raises(ValueError):
        _ = square1._join_via(square2, "invalid")

    assert square1.resample(False) == square1
    assert square1.resample(None).points.shape == square1.points.shape
    assert square1.resample(71).points.shape != square1.points.shape

    with pytest.raises(ValueError):
        bowtie = [(0, 0), (0, 2), (1, 1), (2, 2), (2, 0), (1, 1), (0, 0)]
        _ = tdgl.Polygon(name="bowtie", points=bowtie)

    for min_points, smooth in [(None, None), (500, None), (500, 10)]:
        mesh = square1.make_mesh(min_points=min_points, smooth=smooth)
        if min_points:
            assert mesh.sites.shape[0] > min_points


def test_plot_polygon():
    with tdgl.non_gui_backend():
        ax = tdgl.Polygon("square1", points=tdgl.geometry.box(1)).plot()
        assert isinstance(ax, plt.Axes)
    plt.close("all")


@pytest.fixture(scope="module")
def device():

    layer = tdgl.Layer(london_lambda=1, coherence_length=1, thickness=0.1, z0=0)
    film = tdgl.Polygon("ring", points=tdgl.geometry.ellipse(3, 2, angle=5))
    hole = tdgl.Polygon("hole", points=tdgl.geometry.circle(1))

    offset_film = film.buffer(1, join_style="mitre", as_polygon=False)
    assert isinstance(offset_film, np.ndarray)
    assert offset_film.shape[0] >= film.points.shape[0]

    offset_poly = film.buffer(1)
    assert isinstance(offset_poly, tdgl.Polygon)
    assert film.name in offset_poly.name

    assert film.contains_points([0, 0])
    assert np.array_equal(
        film.contains_points([[0, 0], [2, 1]], index=True), np.array([0, 1])
    )
    assert np.isclose(film.area, np.pi * 3 * 2, rtol=1e-3)

    abstract_regions = [
        tdgl.Polygon(
            "abstract",
            points=tdgl.geometry.box(2.5, angle=45),
        ),
    ]

    bad_probe_points = [
        [(0, 0, 0)],
        [(5, 0), (5, 5)],
    ]

    for probe_points in bad_probe_points:
        with pytest.raises(ValueError):
            _ = tdgl.Device(
                "device",
                layer=layer,
                film=film,
                holes=[hole],
                abstract_regions=abstract_regions,
                probe_points=probe_points,
            )

    device = tdgl.Device(
        "device",
        layer=layer,
        film=film,
        holes=[hole],
        abstract_regions=abstract_regions,
        probe_points=[(-1.5, 0), (1.5, 0)],
        length_units="um",
    )

    with pytest.raises(TypeError):
        device.scale(xfact=-1, origin="center")
    with pytest.raises(TypeError):
        device.rotate(90, origin="centroid")

    assert isinstance(device.scale(xfact=-1), tdgl.Device)
    assert isinstance(device.scale(yfact=-1), tdgl.Device)
    assert isinstance(device.rotate(90), tdgl.Device)
    dx = 1
    dy = -1
    dz = 1
    assert isinstance(device.translate(dx, dy, dz=dz), tdgl.Device)
    d = device.copy()
    assert d.translate(dx, dy, dz=dz, inplace=True) is d

    assert device.layer.conductivity is None
    assert device.conductivity is None
    with pytest.raises(ValueError):
        _ = device.tau0()
    with pytest.raises(ValueError):
        _ = device.V0()

    device.layer.conductivity = 10  # S/um
    assert device.conductivity.to("S/m").magnitude == 10**7
    assert isinstance(device.tau0(), pint.Quantity)
    assert isinstance(device.V0(), pint.Quantity)

    return device


@pytest.fixture(scope="module")
def device_with_mesh():

    with pytest.raises(ValueError):
        _ = tdgl.Polygon("poly", points=tdgl.geometry.circle(1).T)

    layer = tdgl.Layer(london_lambda=1, coherence_length=1, thickness=0.1, z0=0)

    film = tdgl.Polygon(
        "ring",
        points=tdgl.geometry.close_curve(tdgl.geometry.circle(4)),
    )

    holes = [
        tdgl.Polygon("ring_hole", points=tdgl.geometry.circle(2)),
    ]

    device = tdgl.Device(
        "device",
        layer=layer,
        film=film,
        holes=holes,
        probe_points=[(2.5, 0), (-2.5, 0)],
    )
    assert device.edge_lengths is None
    assert device.triangles is None
    assert device.edges is None
    device.make_mesh(min_points=3000)
    assert isinstance(device.edge_lengths, np.ndarray)
    assert isinstance(device.triangles, np.ndarray)
    centroids = device.points[device.triangles].mean(axis=1)
    assert centroids.shape[0] == device.triangles.shape[0]

    print(device)
    assert device == device
    assert device.layer == layer
    assert device.film == film
    assert film != layer
    assert layer != film
    assert layer == layer.copy()
    assert layer is not layer.copy()
    assert film == film.copy()
    assert film is not film.copy()
    assert device != layer

    assert copy.deepcopy(device) == copy.copy(device) == device.copy() == device

    d = device.scale(xfact=-1)
    assert isinstance(d, tdgl.Device)
    assert d.points is None
    d = device.scale(yfact=-1)
    assert isinstance(d, tdgl.Device)
    assert d.points is None
    d = device.rotate(90)
    assert isinstance(d, tdgl.Device)
    assert d.points is None
    dx = 1
    dy = -1
    dz = 1
    assert isinstance(device.translate(dx, dy, dz=dz), tdgl.Device)
    d = device.copy()
    assert d.translate(dx, dy, dz=dz, inplace=True) is d

    for points in ("poly_points", "points"):
        x0, y0 = getattr(device, points).mean(axis=0)
        z0 = device.layer.z0
        with device.translation(dx, dy, dz=dz):
            x, y = getattr(device, points).mean(axis=0)
            assert np.isclose(x, x0 + dx)
            assert np.isclose(y, y0 + dy)
            assert np.isclose(layer.z0, z0 + dz)
        x, y = getattr(device, points).mean(axis=0)
        assert np.isclose(x, x0)
        assert np.isclose(y, y0)
        assert np.isclose(layer.z0, z0)

    return device


@pytest.mark.parametrize("legend", [False, True])
def test_plot_device(
    device: tdgl.Device, device_with_mesh: tdgl.Device, legend, mesh=True
):
    with tdgl.non_gui_backend():
        fig, axes = device.plot(legend=legend)
        with pytest.raises(RuntimeError):
            _ = device.plot(legend=legend, mesh=mesh)
        fig, axes = device_with_mesh.plot(legend=legend, mesh=mesh)
        plt.close("all")


@pytest.mark.parametrize("legend", [False, True])
@pytest.mark.parametrize("exclude", [None, "ring", ["ring"]])
def test_draw_device(device: tdgl.Device, legend, exclude):
    with tdgl.non_gui_backend():
        fig, axes = device.draw(exclude=exclude, legend=legend)
        fig, axes = device.draw(
            legend=legend,
        )

    with tdgl.non_gui_backend():
        fig, ax = plt.subplots()
        _ = tdgl.Device(
            "device",
            layer=tdgl.Layer(london_lambda=1, coherence_length=1, thickness=0.1, z0=0),
            film=tdgl.Polygon("disk", points=tdgl.geometry.circle(1)),
        ).draw(ax=ax)
        plt.close("all")


@pytest.mark.parametrize("show_sites", [False, True])
@pytest.mark.parametrize("show_edges", [False, True])
@pytest.mark.parametrize("show_dual_edges", [False, True])
@pytest.mark.parametrize("show_voronoi_centroids", [False, True])
@pytest.mark.parametrize("ax", [None, "new"])
def test_plot_mesh(
    device_with_mesh: tdgl.Device,
    show_sites,
    show_edges,
    show_dual_edges,
    show_voronoi_centroids,
    ax,
):
    mesh = device_with_mesh.mesh
    kwargs = dict(
        show_sites=show_sites,
        show_edges=show_edges,
        show_dual_edges=show_dual_edges,
        show_voronoi_centroids=show_voronoi_centroids,
    )
    with tdgl.non_gui_backend():
        if ax is not None:
            fig, ax = plt.subplots()
            kwargs["ax"] = ax
        ax = mesh.plot(**kwargs)
        assert isinstance(ax, plt.Axes)
    plt.close("all")


@pytest.mark.parametrize(
    ", ".join(["min_points", "smooth"]),
    [(None, None), (None, 20), (1200, None), (1200, 20)],
)
def test_make_mesh(device: tdgl.Device, min_points, smooth):
    device.make_mesh(
        min_points=min_points,
        smooth=smooth,
    )

    assert device.points is not None
    assert device.triangles is not None
    if min_points:
        assert device.points.shape[0] >= min_points


@pytest.mark.parametrize("save_mesh", [False, True])
def test_device_to_file(device: tdgl.Device, device_with_mesh: tdgl.Device, save_mesh):

    with tempfile.TemporaryDirectory() as directory:
        path = os.path.join(directory, "device.h5")
        device.to_hdf5(os.path.join(path), save_mesh=save_mesh)
        loaded_device = tdgl.Device.from_hdf5(path)
    assert device == loaded_device

    with tempfile.TemporaryDirectory() as directory:
        path = os.path.join(directory, "device.h5")
        device_with_mesh.to_hdf5(path, save_mesh=save_mesh)
        loaded_device = tdgl.Device.from_hdf5(path)
    assert device_with_mesh == loaded_device


def test_pickle_device(device: tdgl.Device, device_with_mesh: tdgl.Device):

    loaded_device = pickle.loads(pickle.dumps(device))
    loaded_device_with_mesh = pickle.loads(pickle.dumps(device_with_mesh))

    assert loaded_device == device
    assert loaded_device_with_mesh == device_with_mesh

    assert loaded_device.ureg("1 m") == loaded_device.ureg("1000 mm")


@pytest.mark.parametrize("origin", [(0, 0), (4, 5)])
def test_transforms(transport_device: tdgl.Device, origin):

    device = transport_device
    device.scale(xfact=2, yfact=0.5, origin=origin)
    device.rotate(-40, origin=origin)
    device.translate(dx=-10, dy=10)

    with device.translation(dx=5, dy=-10):
        device == device.translate(dx=5, dy=-10)



import pytest
from IPython.display import HTML

from tdgl.about import version_dict, version_table


def test_version_dict():
    d = version_dict()
    assert isinstance(d, dict)


@pytest.mark.parametrize("verbose", [False, True])
def test_version_table(verbose):
    html = version_table(verbose=verbose)
    assert isinstance(html, HTML)

    verion_info = version_dict()
    html = version_table(version_info=verion_info, verbose=verbose)
    assert isinstance(html, HTML)



from .about import version_dict, version_table
from .device.device import Device
from .device.layer import Layer
from .device.meshing import generate_mesh
from .device.polygon import Polygon
from .em import ureg
from .fluxoid import Fluxoid, make_fluxoid_polygons
from .parameter import Constant, Parameter
from .solution.plot_solution import (
    plot_currents,
    plot_field_at_positions,
    plot_order_parameter,
    plot_scalar_potential,
    plot_vorticity,
)
from .solution.solution import BiotSavartField, Solution
from .solver.options import SolverOptions
from .solver.solve import solve
from .version import __version__
from .visualization.common import non_gui_backend



import logging
import os
from contextlib import nullcontext
from logging import Logger
from typing import Any, Dict, Sequence, Union

import h5py
import numpy as np
from matplotlib import animation
from matplotlib import pyplot as plt
from matplotlib.ticker import FuncFormatter
from tqdm import tqdm

from ..finite_volume.mesh import Mesh
from ..solution.data import get_data_range
from .common import PLOT_DEFAULTS, Quantity, auto_grid
from .io import get_plot_data, get_state_string


def create_animation(
    input_file: Union[str, h5py.File],
    *,
    output_file: Union[str, None] = None,
    quantities: Union[str, Sequence[str]],
    fps: int = 30,
    dpi: float = 100,
    max_cols: int = 4,
    min_frame: int = 0,
    max_frame: int = -1,
    quiver: bool = False,
    axes_off: bool = False,
    title_off: bool = False,
    full_title: bool = True,
    logger: Union[Logger, None] = None,
    silent: bool = False,
    figure_kwargs: Union[Dict[str, Any], None] = None,
    writer: Union[str, animation.MovieWriter, None] = None,
) -> animation.FuncAnimation:
    """Generates, and optionally saves, and animation of a TDGL simulation.

    The animation will be in dimensionless units.

    Args:
        input_file: An open h5py file or a path to an H5 file containing
            the :class:`tdgl.Solution` you would like to animate.
        output_file: A path to which to save the animation,
            e.g., as a gif or mp4 video.
        quantities: The names of the quantities to animate.
        fps: Frame rate in frames per second.
        dpi: Resolution in dots per inch.
        max_cols: The maxiumum number of columns in the subplot grid.
        min_frame: The first frame of the animation.
        max_frame: The last frame of the animation.
        quiver: Add quiver arrows to the plots.
        axes_off: Turn off the axes for each subplot.
        title_off: Turn off the figure suptitle.
        full_title: Include the full "state" for each frame in the figure suptitle.
        figure_kwargs: Keyword arguments passed to ``plt.subplots()`` when creating
            the figure.
        writer: A :class:`matplotlib.animation.MovieWriter` instance to use when
            saving the animation.
        logger: A logger instance to use.
        silent: Disable logging.

    Returns:
        The animation as a :class:`matplotlib.animation.FuncAnimation`.
    """
    if isinstance(input_file, str):
        input_file = input_file
    if quantities is None:
        quantities = Quantity.get_keys()
    if isinstance(quantities, str):
        quantities = [quantities]
    quantities = [Quantity.from_key(name.upper()) for name in quantities]
    num_plots = len(quantities)
    logger = logger or logging.getLogger()
    figure_kwargs = figure_kwargs or dict()
    figure_kwargs.setdefault("constrained_layout", True)
    default_figsize = (
        3.25 * min(max_cols, num_plots),
        3 * max(1, num_plots // max_cols),
    )
    figure_kwargs.setdefault("figsize", default_figsize)

    logger.info(f"Creating animation for {[obs.name for obs in quantities]!r}.")

    mpl_context = nullcontext() if output_file is None else plt.ioff()
    if isinstance(input_file, str):
        h5_context = h5py.File(input_file, "r", libver="latest")
    else:
        h5_context = nullcontext(input_file)

    with h5_context as h5file:
        with mpl_context:
            # Get the mesh
            if "mesh" in h5file:
                mesh = Mesh.from_hdf5(h5file["mesh"])
            else:
                mesh = Mesh.from_hdf5(h5file["solution/device/mesh"])

            x, y = mesh.sites.T

            # Get the ranges for the frame
            _min_frame, _max_frame = get_data_range(h5file)
            min_frame = max(min_frame, _min_frame)
            if max_frame == -1:
                max_frame = _max_frame
            else:
                max_frame = min(max_frame, _max_frame)

            # Temp data to use in plots
            temp_value = np.ones(len(mesh.sites), dtype=float)
            temp_value[0] = 0
            temp_value[1] = 0.5

            fig, axes = auto_grid(num_plots, max_cols=max_cols, **figure_kwargs)
            collections = []
            for quantity, ax in zip(quantities, axes.flat):
                opts = PLOT_DEFAULTS[quantity]
                collection = ax.tripcolor(
                    x,
                    y,
                    temp_value,
                    triangles=mesh.elements,
                    shading="gouraud",
                    cmap=opts.cmap,
                )
                if quiver:
                    quiver = ax.quiver(
                        x,
                        y,
                        temp_value,
                        temp_value,
                        scale=0.05,
                        units="dots",
                    )
                cbar = fig.colorbar(
                    collection, ax=ax, format=FuncFormatter("{:.2f}".format)
                )
                cbar.set_label(opts.clabel)
                ax.set_aspect("equal")
                ax.set_title(quantity.value)
                if axes_off:
                    ax.axis("off")
                collections.append(collection)

            vmins = [+np.inf for _ in quantities]
            vmaxs = [-np.inf for _ in quantities]

            def update(frame):
                if not h5file:
                    return
                state = get_state_string(h5file, frame, max_frame)
                if not full_title:
                    state = state.split(",")[0]
                if not title_off:
                    fig.suptitle(state)
                for i, (quantity, collection) in enumerate(
                    zip(quantities, collections)
                ):
                    value, direction, limits = get_plot_data(
                        h5file, mesh, quantity, frame
                    )
                    vmins[i] = min(vmins[i], limits[0])
                    vmaxs[i] = max(vmaxs[i], limits[1])
                    collection.set_array(value)
                    collection.set_clim(vmins[i], vmaxs[i])
                if quiver:
                    quiver.set_UVC(direction[:, 0], direction[:, 1])
                fig.canvas.draw()

            anim = animation.FuncAnimation(
                fig,
                update,
                frames=max_frame - min_frame,
                interval=1e3 / fps,
                blit=False,
            )

        if output_file is not None:
            output_file = os.path.join(os.getcwd(), output_file)
            if writer is None:
                kwargs = dict(fps=fps)
            else:
                kwargs = dict(writer=writer)
            fname = os.path.basename(output_file)
            with tqdm(
                total=len(range(min_frame, max_frame)),
                unit="frames",
                disable=silent,
                desc=f"Saving to {fname}",
            ) as pbar:
                anim.save(
                    output_file,
                    dpi=dpi,
                    progress_callback=lambda frame, total: pbar.update(1),
                    **kwargs,
                )

        return anim



import os
import subprocess
import tempfile
from typing import List

import matplotlib.pyplot as plt
import pytest

import tdgl
from tdgl import visualize
from tdgl.visualization import DEFAULT_QUANTITIES, Quantity, create_animation


@pytest.fixture(scope="module")
def tempdir():
    tmp = tempfile.TemporaryDirectory()
    yield tmp.__enter__()
    tmp.cleanup()


@pytest.fixture(scope="module")
def solution(transport_device, tempdir):
    device = transport_device
    total_time = 10

    fname = os.path.join(tempdir, "output.h5")
    options = tdgl.SolverOptions(
        solve_time=total_time,
        output_file=fname,
        field_units="uT",
        current_units="uA",
    )
    field = tdgl.sources.ConstantField(10)
    solution = tdgl.solve(
        device,
        options,
        applied_vector_potential=field,
        terminal_currents=dict(source=10, drain=-10),
    )
    return solution


def run_cmd(cmd: List[str]) -> None:
    result = subprocess.run(
        cmd,
        check=True,
        capture_output=True,
        text=True,
    )
    print(result.stderr)
    print(result.stdout)


def test_visualize_help():
    run_cmd(["python", "-m", "tdgl.visualize", "-h"])


def test_no_args():
    with pytest.raises(subprocess.CalledProcessError):
        run_cmd(["python", "-m", "tdgl.visualize"])


def test_bad_input():
    with pytest.raises(subprocess.CalledProcessError):
        run_cmd(["python", "-m", "tdgl.visualize"])


@pytest.mark.parametrize(
    "quantities", [None, "all"] + [name.lower() for name in Quantity.get_keys()]
)
@pytest.mark.parametrize("silent", [False, True])
@pytest.mark.parametrize("verbose", [False, True])
def test_interactive(solution, quantities, verbose, silent):
    parser = visualize.make_parser()
    cmd = ["--input", solution.path, "interactive"]
    if verbose:
        cmd.insert(2, "--verbose")
    if silent:
        cmd.insert(2, "--silent")
    if quantities is not None:
        cmd.extend(["--quantities"] + quantities.split(" "))
    args = parser.parse_args(cmd)
    with tdgl.non_gui_backend():
        tdgl.visualize.main(args)
        plt.close("all")


@pytest.mark.parametrize("quantities", [None, "order_parameter phase", "supercurrent"])
@pytest.mark.parametrize("ext", ["-m.gif"])
@pytest.mark.parametrize("max_cols", [4, 2])
def test_animation(solution, quantities, ext, max_cols):
    kwargs = dict(
        output_file=solution.path.replace(".h5", ext),
        full_title=False,
        dpi=200,
        fps=20,
        max_cols=max_cols,
    )
    if quantities is None:
        kwargs["quantities"] = DEFAULT_QUANTITIES
    if quantities is not None:
        kwargs["quantities"] = quantities.split(" ")
    create_animation(solution.path, **kwargs)


@pytest.mark.parametrize(
    "quantities",
    [
        "all",
        # None,
        # "order_parameter phase vorticity supercurrent",
    ],
)
def test_animate_cli(solution, quantities):
    parser = visualize.make_parser()
    cmd = [
        "--input",
        solution.path,
        "animate",
        "--output",
        solution.path.replace(".h5", "-cli.gif"),
        "--min-frame",
        "2",
        "--max-frame",
        "-1",
        "--fps",
        "30",
        "--dpi",
        "200",
    ]
    if quantities is not None:
        cmd.extend(["--quantities"] + quantities.split(" "))
    args = parser.parse_args(cmd)
    with tdgl.non_gui_backend():
        tdgl.visualize.main(args)
        plt.close("all")
